from __future__ import absolute_import

import importlib
import os
import tldextract

from lexicon.config import ConfigResolver, DictConfigSource
from lexicon.config import non_interactive_config_resolver, legacy_config_resolver

# From providers import Example
class Client(object):
    def __init__(self, config = None):
        if not config:
            # If there is not config specified, we load a non-interactive configuration.
            self.config = non_interactive_config_resolver()
        elif not isinstance(config, ConfigResolver):
            # If config is not a ConfigResolver, we are in a legacy situation.
            # We protect this part of the Client API.
            self.config = legacy_config_resolver(config)
        else:
            self.config = config

        # Validate configuration
        self._validate_config()

        runtime_config = {}

        # Process domain, strip subdomain
        domain_parts = tldextract.extract(self.config.resolve('lexicon:domain'))
        runtime_config['domain'] = '{0}.{1}'.format(domain_parts.domain, domain_parts.suffix)

        if self.config.resolve('lexicon:delegated'):
            # handle delegated domain
            delegated = self.config.resolve('lexicon:delegated').rstrip('.')
            if delegated != runtime_config.get('domain'):
                # convert to relative name
                if delegated.endswith(runtime_config.get('domain')):
                    delegated = delegated[:-len(runtime_config.get('domain'))]
                    delegated = delegated.rstrip('.')
                # update domain
                runtime_config['domain'] = '{0}.{1}'.format(delegated, runtime_config.get('domain'))

        self.action = self.config.resolve('lexicon:action')
        self.provider_name = self.config.resolve('lexicon:provider_name') or self.config.resolve('lexicon:provider')

        self.config.add_config_source(DictConfigSource(runtime_config), 0)

        provider_module = importlib.import_module('lexicon.providers.' + self.provider_name)
        provider_class = getattr(provider_module, 'Provider')
        self.provider = provider_class(self.config)

    def execute(self):
        self.provider.authenticate()
        identifier = self.config.resolve('lexicon:identifier')
        record_type = self.config.resolve('lexicon:type')
        name = self.config.resolve('lexicon:name')
        content = self.config.resolve('lexicon:content')

        if self.action == 'create':
            return self.provider.create_record(record_type, name, content)

        elif self.action == 'list':
            return self.provider.list_records(record_type, name, content)

        elif self.action == 'update':
            return self.provider.update_record(identifier, record_type, name, content)

        elif self.action == 'delete':
            return self.provider.delete_record(identifier, record_type, name, content)

    def _validate_config(self):
        if not self.config.resolve('lexicon:provider_name'):
            raise AttributeError('provider_name')
        if not self.config.resolve('lexicon:action'):
            raise AttributeError('action')
        if not self.config.resolve('lexicon:domain'):
            raise AttributeError('domain')
        if not self.config.resolve('lexicon:type'):
            raise AttributeError('type')



from __future__ import absolute_import

import json
import logging
import time
import requests
import binascii

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding
from base64 import urlsafe_b64encode, b64decode

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['googledomains.com']

# Implements the Google Cloud DNS provider.
# This API is quite complicated to use, as it used some unique concepts compared to other providers.
# First of all, it uses a full-fledged OAuth2 authentication, involving signing a JWT and retrieve a Bearer token.
# This hard work is done in the authenticate() process, using the strong and well known "cryptography" package.
# Second, Google Cloud DNS API contains this really particular patterns:
#   - all records of the same type and name are stacked together in a RecordSet representation, 
#       which contains in the rrdatas array all current values for this type/name pair, including
#       explictly monovalued entries like A or CNAME.
#   - modifications can only done through a create/delete pattern: there is no way to update a record
#   - more importantly, this approach extends to all values of a given type/name pair: it means that adding/removing
#       a new value to a TXT entry requires to delete all values of this entry, then recreate it with all 
#       values desired (the old ones plus the new one for adding, the old ones minus the removed one for removing)
# So all the hard work in this provider, appart from the authentication process, is to convert the Lexicon monovalued
#   entries representation to/from the Google multivalued and stacked representation 
#   through create/update/list/delete processes.
def ProviderParser(subparser):
    subparser.description = '''
        The Google Cloud DNS provider requires the JSON file which contains the service account info to connect to the API.
        This service account must own the project role DNS > DNS administrator for the project associated to the DNS zone.
        You can create a new service account, associate a private key, and download its info through this url: 
        https://console.cloud.google.com/iam-admin/serviceaccounts?authuser=2'''
    subparser.add_argument('--auth-service-account-info', help='''
        specify the service account info in the Google JSON format: 
        can be either the path of a file prefixed by 'file::' (eg. file::/tmp/service_account_info.json)
        or the base64 encoded content of this file prefixed by 'base64::' (eg. base64::eyJhbGciOyJ...)''')

class Provider(BaseProvider):

    # We need serveral parameters, which are available in the JSON file provided 
    #   by Google when associating a private key to the relevant service account.
    # So this JSON file is the natural input to configure the provider.
    # It can be provided as a path to the JSON file, or as its content encoded
    #   in base64, which is a suitable portable way in particular for Docker containers.
    # In both cases the content is loaded as bytes, on loaded in a private instance variable.
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self._token = None

        if self._get_provider_option('auth_service_account_info').startswith('file::'):
            with open(self._get_provider_option('auth_service_account_info').replace('file::', ''), 'rb') as file:
                service_account_info_bytes = file.read()
        elif self._get_provider_option('auth_service_account_info').startswith('base64::'):
            service_account_info_bytes = b64decode(self._get_provider_option('auth_service_account_info').replace('base64::', ''))
        else:
            raise Exception('Invalid value passed to --auth-service-account-info, should be a path prefixed with \'file::\' or a base64 value prefixed by \'base64::\'.')

        self._service_account_info = json.loads(service_account_info_bytes.decode('utf-8'))

        if not self._service_account_info['client_email'] or not self._service_account_info['private_key'] or not self._service_account_info['project_id']:
            raise Exception('Invalid service account info (missing either client_email/private_key/project_id key).')
    
    # We have a real authentication here, which uses the OAuth protocol:
    #   - a JWT token is forged with the Google Cloud DNS access claims, using the service account info loaded by the constructor,
    #   - this JWT token is signed by a PKCS1v15 signature using the RSA private key associated to the service account,
    #   - this JWT token is then submitted to the Google API, which returns an access token
    #   - this access token will be used for every future HTTP request to the Google Cloud DNS API to authenticate the user.
    #   - finally we make a first authenticated request to retrieve the managed zone id, which will also be used on future requests.
    # This access token has a default lifetime of 10 minutes, but is used only for the current Lexicon operation, so it should be sufficient.
    def authenticate(self):
        jwt_header = {
            'alg': 'RS256',
            'typ': 'JWT'
        }
        jwt_header_bytes = urlsafe_b64encode(json.dumps(jwt_header).encode('utf-8'))
        
        epoch_time = int(time.time())
        jwt_claims_set = {
            'iss': self._service_account_info['client_email'],
            'scope': 'https://www.googleapis.com/auth/ndev.clouddns.readwrite',
            'aud': 'https://www.googleapis.com/oauth2/v4/token',
            'exp': epoch_time + 60 * 10,
            'iat': epoch_time
        }
        jwt_claims_set_bytes = urlsafe_b64encode(json.dumps(jwt_claims_set).encode('utf-8'))

        private_key = serialization.load_pem_private_key(
            self._service_account_info['private_key'].encode('utf-8'),
            password=None,
            backend=default_backend()
        )

        jwt_sign = private_key.sign(
            b'.'.join([jwt_header_bytes, jwt_claims_set_bytes]),
            padding.PKCS1v15(),
            hashes.SHA256()
        )
        jwt_sign_bytes = urlsafe_b64encode(jwt_sign)

        jwt_bytes = b'.'.join([jwt_header_bytes, jwt_claims_set_bytes, jwt_sign_bytes])

        post_data = {
            'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
            'assertion': jwt_bytes
        }
        post_header = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }

        auth_request = requests.request('POST', 'https://www.googleapis.com/oauth2/v4/token', data=post_data, headers=post_header)

        auth_request.raise_for_status()
        post_result = auth_request.json()

        if not post_result['access_token']:
            raise Exception('Error, could not grant RW access on the Google Cloud DNS API for user: {0}'.format(self._get_provider_option('auth_email')))

        self._token = post_result['access_token']

        results = self._get('/managedZones')

        targetedManagedZoneIds = [managedZone['id'] for managedZone in results['managedZones'] if managedZone['dnsName'] == '{0}.'.format(self.domain)]

        if not targetedManagedZoneIds:
            raise Exception('Error, domain {0} is not registered for this project'.format(self.domain))

        self.domain_id = targetedManagedZoneIds[0]

    # List all records for the given type/name/content.
    # It is quite straight forward to request data, the biggest operation is to convert
    #   the stacked multivalued RecordSets into Lexicon monovalued entries.
    # Plese not that we could provide type and name to the API to make the filtering, 
    #   but providing the type makes the name mandatory with the Google Cloud DNS API, and
    #   name is not always available (we can ask for every TXT record for example). So to stick to
    #   the most general case, its preferable to always get all records and be free to filter
    #   the way we want afterwards.
    def list_records(self, type=None, name=None, content=None):
        results = self._get('/managedZones/{0}/rrsets'.format(self.domain_id))

        records = []

        for rrset in results['rrsets']:
            for rrdata in rrset['rrdatas']:
                record = {
                    'type': rrset['type'],
                    'name': self._full_name(rrset['name']),
                    'ttl': rrset['ttl'],
                    'content': rrdata
                }
                self._clean_TXT_record(record)
                record['id'] = Provider._identifier(record)
                records.append(record)

        if type:
            records = [record for record in records if record['type'] == type]
        if name:
            records = [record for record in records if record['name'] == self._full_name(name)]
        if content:
            records = [record for record in records if record['content'] == content]

        LOGGER.debug('list_records: %s', records)

        return records

    # Create the record with provided type, name and content.
    # Because of the way this API is constructed, it is quite complex in fact.
    # Indeed we need to know if there is already a RecordSet for the type/name pair, and update
    #   or create accordingly the RecordSet. Furthermore, we need first to delete the old RecordSet
    #   if it exists, to replace it with the RecordSet containing the new content we want.
    def create_record(self, type, name, content):
        if not type or not name or not content:
            raise Exception('Error, type, name and content are mandatory to create a record.')

        identifier = Provider._identifier({'type': type, 'name': self._full_name(name), 'content': content})

        query_params = {
            'type': type,
            'name': self._fqdn_name(name)
        }

        results = self._get('/managedZones/{0}/rrsets'.format(self.domain_id), query_params=query_params)

        rrdatas = []
        changes = {}
        if results['rrsets']:
            rrset = results['rrsets'][0]
            for rrdata in rrset['rrdatas']:
                if rrdata == Provider._normalize_content(rrset['type'], content):
                    LOGGER.debug('create_record (ignored, duplicate): %s', identifier)
                    return True

            changes['deletions'] = [{
                'name': rrset['name'],
                'type': rrset['type'],
                'ttl': rrset['ttl'],
                'rrdatas': rrset['rrdatas'][:]
            }]

            rrdatas = rrset['rrdatas'][:]
        
        rrdatas.append(Provider._normalize_content(type, content))

        changes['additions'] = [{
            'name': self._fqdn_name(name),
            'type': type,
            'ttl': self._get_lexicon_option('ttl'),
            'rrdatas': rrdatas
        }]

        self._post('/managedZones/{0}/changes'.format(self.domain_id), data=changes)

        LOGGER.debug('create_record: %s', identifier)

        return True

    # Update a record for the given identifier or type/name pair with the given content if provided.
    # Again because of the API specification, updating is even more complex than creating, as we need
    #   to take into account every RecordSet that should be destroyed then recreated.
    # As all the hard work has been done on list_record, create_record and delete_record, we use a
    #   combination of these three methods to obtain the state we want.
    # Even if this make the operation very costly regarding the number of requests to do, it allows
    #   the implementation to be way more readable (without that, it would take grossly the size of 
    #   the three quoted methods).
    def update_record(self, identifier, type=None, name=None, content=None):
        if not identifier and (not type or not name):
            raise Exception('Error, identifier or type+name parameters are required.')

        if identifier:
            records = self.list_records()
            records_to_update = [record for record in records if record['id'] == identifier]
        else:
            records_to_update = self.list_records(type=type, name=name)

        if not records_to_update:
            raise Exception('Error, could not find a record for given identifier: {0}'.format(identifier))

        if len(records_to_update) > 1:
            LOGGER.warn('Warning, multiple records found for given parameters, only first one will be updated: %s', records_to_update)

        record_identifier = records_to_update[0]['id']

        original_level = LOGGER.getEffectiveLevel()
        LOGGER.setLevel(logging.WARNING)
        self.delete_record(record_identifier)

        new_record = {
            'type': type if type else records_to_update[0]['type'], 
            'name': name if name else records_to_update[0]['name'], 
            'content': content if content else records_to_update[0]['content']
        }

        self.create_record(new_record['type'], new_record['name'], new_record['content'])
        LOGGER.setLevel(original_level)

        LOGGER.debug('update_record: %s => %s', record_identifier, Provider._identifier(new_record))

        return True

    # Delete a record for the given identifier or the given type/name/content.
    # Really complex to do, because a lot of RecordSets can be updated (so destroyed and recreated)
    #   depending on the given condition (eg. with content alone, every record could be inspected).
    # There is mainly to cases:
    #   - either an association of one or more between type, name and content is given
    #   - either an identifier is given, and we extract the type + name + content to process as the first case.
    # Anyway, we will need to parse every RecordSet available, and for each of them which match the conditions:
    #   - mark as deletion the existing RecordSet
    #   - remove the targeted content from the RecordSet
    #   - mark as addition the update RecordSet with the subset of rrdatas if rrdatas is not empty
    #   - do not mark as additions RecordSets whose rrdatas subset become empty: 
    #       for this type/name pair, all RecordSet needs to go away.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        results = self._get('/managedZones/{0}/rrsets'.format(self.domain_id))

        if identifier:
            changes = self._process_records_to_delete_by_identifier(results, identifier)
        else:
            changes = self._process_records_to_delete_by_parameters(results, type, name, content)

        if not changes:
            raise Exception('Could not find existing record matching the given parameters.')

        self._post('/managedZones/{0}/changes'.format(self.domain_id), data=changes)

        LOGGER.debug('delete_records: %s %s %s %s', identifier, type, name, content)

        return True

    # Calculate the changes to do based on the record to remove identified by its identifier.
    # This implementation find the corresponding record, and use its type + name + value to
    #   delegate the processing to _process_records_to_delete_by_parameters.
    def _process_records_to_delete_by_identifier(self, results, identifier):
        for rrset in results['rrsets']:
            for rrdata in rrset['rrdatas']:
                record = {
                    'type': rrset['type'], 
                    'name': self._full_name(rrset['name']), 
                    'content': rrdata
                }

                self._clean_TXT_record(record)
                record_identifier = Provider._identifier(record)

                if identifier == record_identifier:
                    return self._process_records_to_delete_by_parameters(results, record['type'], record['name'], record['content'])

        return None

    # Calculate the changes to do based on the records to remove identified by type/name/content.
    # Additions and deletions are registered accordingly in the changes, and RecordSet with empty
    #   rrdatas after its subset are not marked in additions to be completly removed from the DNS zone.
    def _process_records_to_delete_by_parameters(self, results, type=None, name=None, content=None):
        rrsets_to_modify = results['rrsets']

        if type:
            rrsets_to_modify = [rrset for rrset in rrsets_to_modify if rrset['type'] == type]
        if name:
            rrsets_to_modify = [rrset for rrset in rrsets_to_modify if rrset['name'] == self._fqdn_name(name)]
        if content:
            rrsets_to_modify = [rrset for rrset in rrsets_to_modify if ('"{0}"'.format(content) if rrset['type'] == 'TXT' else content) in rrset['rrdatas']]

        changes = {
            'additions': [], 
            'deletions': []
        }

        for rrset_to_modify in rrsets_to_modify:
            changes['deletions'].append({
                'name': rrset_to_modify['name'],
                'type': rrset_to_modify['type'],
                'ttl': rrset_to_modify['ttl'],
                'rrdatas': rrset_to_modify['rrdatas'][:]
            })

            if content:
                new_rrdatas = rrset_to_modify['rrdatas'][:]
                new_rrdatas.remove('"{0}"'.format(content) if rrset_to_modify['type'] == 'TXT' else content)
                if new_rrdatas:
                    changes['additions'].append({
                        'name': rrset_to_modify['name'],
                        'type': rrset_to_modify['type'],
                        'ttl': rrset_to_modify['ttl'],
                        'rrdatas': new_rrdatas
                    })

        if not changes['additions'] and not changes['deletions']:
            return None

        return changes

    # With Google Cloud DNS API, content of CNAME entries must be FQDN (with a trailing dot),
    #   and content of TXT entries must be quoted. This static method ensures that.
    @staticmethod
    def _normalize_content(type, content):
        if type == 'TXT':
            return '"{0}"'.format(content)
        if type == 'CNAME':
            return '{0}.'.format(content) if not content.endswith('.') else content
        
        return content

    # Google Cloud DNS API does not provide identifier for RecordSets.
    # So we need to calculate our own identifier at runtime.
    # It is based on a SHA256 hash with the most relevant parameters of a record: type, name and content.
    # Note that the identifier is calculated on a Lexicon monovalued entry, not a Google stacked multivalued RecordSet,
    #   to make it usable during Lexicon calls to updates and deletions.
    @staticmethod
    def _identifier(record):
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(('type=' + record.get('type', '') + ',').encode('utf-8'))
        digest.update(('name=' + record.get('name', '') + ',').encode('utf-8'))
        digest.update(('content=' + record.get('content', '') + ',').encode('utf-8'))

        return binascii.hexlify(digest.finalize()).decode('utf-8')[0:7]

    # The request, when authenticated, is really standard:
    #   the request body is encoded as application/json for POST (so the use of 'json' config instead of 'data' in request),
    #   the body response is also encoded as application/json for GET and POST,
    #   and the request headers must contain the access token in the 'Authorization' field.
    def _request(self, action='GET',  url='/', data=None, query_params=None):
        request = requests.request(action, 
                                   'https://content.googleapis.com/dns/v1/projects/{0}{1}'.format(self._service_account_info['project_id'], url), 
                                   params=None if not query_params else query_params,
                                   json=None if not data else data,
                                   headers={'Authorization': 'Bearer {0}'.format(self._token)})

        request.raise_for_status()
        return request.json()



from __future__ import absolute_import

import json
import logging

import requests
from requests.auth import HTTPBasicAuth

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = []

def ProviderParser(subparser):
    subparser.description = '''
        The OnApp provider requires your OnApp account\'s email address and 
        API token, which can be found on your /profile page on the Control Panel interface.
        The server is your dashboard URL, in the format of e.g. https://dashboard.youronapphost.org'''
    subparser.add_argument('--auth-username', help='specify email address of the OnApp account')
    subparser.add_argument('--auth-token', help='specify API Key for the OnApp account')
    subparser.add_argument('--auth-server', help='specify URL to the OnApp Control Panel Server')

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.domain_id = None

        if not self._get_provider_option('auth_username'):
            raise Exception('Error, OnApp Email Address is not defined')
        if not self._get_provider_option('auth_token'):
            raise Exception('Error, OnApp API Key is not defined')
        if not self._get_provider_option('auth_server'):
            raise Exception('Error, OnApp Control Panel URL is not defined')

        self.session = requests.Session()
    
    def authenticate(self):
        domain = self.domain
        
        zones = self._get('/dns_zones.json')
        for zone in zones:
            if zone['dns_zone']['name'] == domain:
                self.domain_id = zone['dns_zone']['id']
                break
        
        if self.domain_id == None:
            raise Exception('Could not find {0} in OnApp DNS Zones'.format(domain))

    def create_record(self, type, name, content):
        data = {
            'name': self._relative_name(name),
            'type': type,
            self._key_for_record_type(type): content
        }

        ttl = self._get_lexicon_option('ttl')
        if ttl:
            data['ttl'] = "{0}".format(ttl)

        result = self._post('/dns_zones/{0}/records.json'.format(self.domain_id), { 'dns_record': data })
        logger.debug('create_record: %s', result)

        return True

    def list_records(self, type=None, name=None, content=None):
        records = []
        
        response = self._get('/dns_zones/{0}/records.json'.format(self.domain_id))
        for recordType in response['dns_zone']['records']:

            # For now we do not support other RR types so we ignore them, also see _key_for_record_type 
            if recordType not in ('A','AAAA','CNAME','TXT'):
                continue

            if type and recordType != type:
                continue

            for record in response['dns_zone']['records'][recordType]:
                record = record['dns_record']

                if name and record['name'] != self._relative_name(name):
                    continue

                recordContent = record[self._key_for_record_type(recordType)]
                if content and recordContent != content:
                    continue

                records.append({
                    'id': record['id'],
                    'name': self._full_name(record['name']),
                    'type': record['type'],
                    'ttl': record['ttl'],
                    'content': recordContent
                })

        logger.debug('list_records: %s', records)

        return records

    def update_record(self, identifier, type=None, name=None, content=None):
        if not identifier:
            existing = self._guess_record(type, name)
            identifier = existing['id']

        ttl = self._get_lexicon_option('ttl')

        if not name or not ttl:
            if not existing:
                existing = self._get('/dns_zones/{0}/records/{1}.json'.format(self.domain_id, identifier))
            if not name:
                name = existing['name']
            if not ttl:
                ttl = existing['ttl']

        request = {
            'name': self._relative_name(name),
            'ttl': '{0}'.format(ttl),
            self._key_for_record_type(type): content
        }

        result = self._put('/dns_zones/{0}/records/{1}.json'.format(self.domain_id, identifier), { 'dns_record': request })
        logger.debug('update_record: %s', result)

        return True

    def delete_record(self, identifier=None, type=None, name=None, content=None):
        deletion_ids = []

        if not identifier:
            records = self.list_records(type, name, content)
            deletion_ids = [ record['id'] for record in records ]
        else:
            deletion_ids.append(identifier)

        for id in deletion_ids:
            self._delete('/dns_zones/{0}/records/{1}.json'.format(self.domain_id, id))

        logger.debug('delete_record: %s', True)

        return True

    def _request(self, action='GET', url='/', data=None, query_params=None):
        headers = {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }
        target = self._get_provider_option('auth_server') + url
        
        body = ''
        if data is not None:
            body = json.dumps(data)

        auth = HTTPBasicAuth(self._get_provider_option('auth_username'), self._get_provider_option('auth_token'))

        request = requests.Request(action, target, data=body, headers=headers, params=query_params, auth=auth)
        prepared_request = self.session.prepare_request(request)

        result = self.session.send(prepared_request)
        result.raise_for_status()

        if result.text:
            return result.json()
        else:
            return None

    def _key_for_record_type(self, record_type):
        if record_type in ('A','AAAA'):
            return 'ip'
        elif record_type == 'CNAME':
            return 'hostname'
        elif record_type == 'TXT':
            return 'txt'
        elif record_type in ('MX','NS', 'SOA', 'SRV', 'LOC'):
            raise Exception('{0} record type is not supported in the OnApp Provider'.format(record_type))

    def _guess_record(self, type, name=None, content=None):
        records = self.list_records(type=type, name=name, content=content)
        if len(records) == 1:
            return records[0]
        elif len(records) > 1:
            raise Exception('Identifier was not provided and several existing records match the request for {0}/{1}'.format(type,name))
        elif len(records) == 0:
            raise Exception('Identifier was not provided and no existing records match the request for {0}/{1}'.format(type,name))    


from __future__ import absolute_import

import json
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['digitalocean.com']

def ProviderParser(subparser):
    subparser.add_argument("--auth-token", help="specify token for authentication")

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.digitalocean.com/v2'

    def authenticate(self):

        payload = self._get('/domains/{0}'.format(self.domain))
        self.domain_id = self.domain

    def create_record(self, type, name, content):
        # check if record already exists
        if len(self.list_records(type, name, content)) == 0:
            record = {
                'type': type,
                'name': self._relative_name(name),
                'data': content,

            }
            if type == 'CNAME':
                record['data'] = record['data'].rstrip('.') + '.' # make sure a the data is always a FQDN for CNAMe.

            payload = self._post('/domains/{0}/records'.format(self.domain_id), record)
        logger.debug('create_record: %s', True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        url = '/domains/{0}/records'.format(self.domain_id)
        records = []
        payload = {}

        next = url
        while next is not None:
            payload = self._get(next)
            if 'links' in payload \
                    and 'pages' in payload['links'] \
                    and 'next' in payload['links']['pages']:
                next = payload['links']['pages']['next']
            else:
                next = None

            for record in payload['domain_records']:
                processed_record = {
                    'type': record['type'],
                    'name': "{0}.{1}".format(record['name'], self.domain_id),
                    'ttl': '',
                    'content': record['data'],
                    'id': record['id']
                }
                records.append(processed_record)

        if type:
            records = [record for record in records if record['type'] == type]
        if name:
            records = [record for record in records if record['name'] == self._full_name(name)]
        if content:
            records = [record for record in records if record['content'].lower() == content.lower()]

        logger.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):

        data = {}
        if type:
            data['type'] = type
        if name:
            data['name'] = self._relative_name(name)
        if content:
            data['data'] = content

        payload = self._put('/domains/{0}/records/{1}'.format(self.domain_id, identifier), data)

        logger.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(type, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)
        
        logger.debug('delete_records: %s', delete_record_id)
        
        for record_id in delete_record_id:
            payload = self._delete('/domains/{0}/records/{1}'.format(self.domain_id, record_id))

        # is always True at this point, if a non 200 response is returned an error is raised.
        logger.debug('delete_record: %s', True)
        return True


    # Helpers
    def _request(self, action='GET',  url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': 'Bearer {0}'.format(self._get_provider_option('auth_token'))
        }
        if not url.startswith(self.api_endpoint):
            url = self.api_endpoint + url

        r = requests.request(action, url, params=query_params,
                             data=json.dumps(data),
                             headers=default_headers)
        r.raise_for_status()  # if the request fails for any reason, throw an error.
        if action == 'DELETE':
            return ''
        else:
            return r.json()



from __future__ import absolute_import

import logging

import requests
from collections import OrderedDict

try:
    import xmltodict # optional dependency
except ImportError:
    pass

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

# Lexicon Plesk Provider
#
# Author: Jens Reimann, 2018
#
# API Docs: https://docs.plesk.com/en-US/onyx/api-rpc
#

plesk_url_suffix = "/enterprise/control/agent.php"

NAMESERVER_DOMAINS = []

def ProviderParser(subparser):
    subparser.add_argument("--auth-username", help="specify username for authentication")
    subparser.add_argument("--auth-password", help="specify password for authentication")
    subparser.add_argument('--plesk-server', help="specify URL to the Plesk Web UI, including the port")

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.api_endpoint = self._get_provider_option('plesk_server')

        if self.api_endpoint.endswith('/'):
            self.api_endpoint = self.api_endpoint[:-1]

        if not self.api_endpoint.endswith(plesk_url_suffix):
            self.api_endpoint += plesk_url_suffix

        self.site_name = self.domain
        assert self.site_name is not None

        self.domain_id = None

        self.username = self._get_provider_option('auth_username')
        assert self.username is not None

        self.password = self._get_provider_option('auth_password')
        assert self.password is not None

    def __simple_request(self, type, operation, req):
        
        response = self.__plesk_request({
            type: {
                operation: req
            }
        })[type][operation]
        
        result = response["result"]
        
        if isinstance(result, list):
            for r in result:
                if r["status"] == "error":
                    raise Exception("API returned at least one error: %s" % r["errtext"] )
        elif response["result"]["status"] == "error":
            errcode = response["result"]["errcode"]
            errtext = response["result"]["errtext"]
            raise Exception("API returned error: %s (%s)" % ( errcode, errtext ) )
        
        return response

    def __plesk_request(self, request):

        headers = {}
        headers["Content-type"] = "text/xml"
        headers["HTTP_PRETTY_PRINT"] = "TRUE"
        headers["HTTP_AUTH_LOGIN"] = self.username
        headers["HTTP_AUTH_PASSWD"] = self.password

        xml = xmltodict.unparse({
                "packet": request
            }, pretty=True)

        logger.debug ( "Request: %s", xml)

        r = requests.post(self.api_endpoint, headers=headers, data=xml,  auth=(self.username, self.password))

        data = r.text

        logger.debug ( "Response: %s", data )
        result = xmltodict.parse(data )
        return result["packet"]

    def __find_site(self):
        return self.__simple_request('site', 'get', OrderedDict([
            ('filter', {'name': self.site_name}),
            ('dataset', {})
        ]))["result"]["id"]

    def authenticate(self):
        self.domain_id = self.__find_site()
        
        if self.domain_id == None:
            raise Exception('Domain not found')

    def create_record(self, type, name, content):
        return self.__create_entry(type, name, content, None)

    def list_records(self, type=None, name=None, content=None):
        entries = self.__find_dns_entries(type, name, content)
        logger.debug("list_records: %s" % entries)
        return entries

    def update_record(self, identifier, type=None, name=None, content=None):
        
        if identifier is None:
            entries = self.__find_dns_entries(type, name, None)
            logger.debug("Entries found: %s", entries)
            
            if len(entries) < 1:
                raise Exception("No entry found for updating")
            
            identifier = entries[0]["id"]
            entry = self.__get_dns_entry(identifier)
            
            ids = []
            for e in entries:
                ids.append(e["id"])
            
            self.__delete_dns_records_by_id(ids)
            
        else:
            
            entry = self.__get_dns_entry(identifier)
            self.__delete_dns_records_by_id([identifier])

        assert entry is not None

        logger.debug("Updating: %s", entry)

        if type:
            entry["type"] = type
        if name:
            entry["host"] = name
        if content:
            entry["value"] = content
        
        return self.__create_entry(entry["type"], entry["host"], entry["value"], entry["opt"])

    def __create_entry(self, type, host, value, opt):
        entries = self.__find_dns_entries(type, self._fqdn_name(host), value)
        
        if entries:
            return True # already exists

        self.__simple_request('dns','add_rec',OrderedDict([
            ('site-id', self.domain_id),
            ('type', type),
            ('host', self._relative_name(host)),
            ('value', value),
            ('opt', opt)
        ]))
        
        return True


    def delete_record(self, identifier=None, type=None, name=None, content=None):
        
        if identifier:
            
            self.__delete_dns_records_by_id([identifier])
            return True
            
        else:
            
            entries = self.__find_dns_entries ( type, self._fqdn_name(name), content )
            ids = []
            
            for entry in entries:
                ids.append(entry["id"])
            
            self.__delete_dns_records_by_id(ids)
            return len(ids) > 0

    def __get_dns_entry(self, id):
        return self.__simple_request('dns', 'get_rec', {
            'filter': {
                'id': id
            }
        })["result"]["data"]

    def __find_dns_entries(self, type = None, host = None, value = None):
        logger.debug("Searching for: %s, %s, %s", type, host, value )

        if value and type and type in ["CNAME"]:
            logger.debug("CNAME transformation")
            value = value.rstrip('.') + "."

        if host:
            host = self._fqdn_name(host)

        result = self.__simple_request('dns', 'get_rec', {
           'filter': {
                'site-id': self.domain_id
            }
        })

        entries = []

        for r in result["result"]:

            logger.debug("Record: %s", r)

            if ( type is not None ) and ( r["data"]["type"] != type ):
                logger.debug("\tType doesn't match - expected: '%s', found: '%s'", type, r["data"]["type"])
                continue
            
            if ( host is not None ) and ( r["data"]["host"] != host ):
                logger.debug("\tHost doesn't match - expected: '%s', found: '%s'", host, r["data"]["host"])
                continue
            
            if ( value is not None ) and ( r["data"]["value"] != value ):
                logger.debug("\tValue doesn't match - expected: '%s', found: '%s'", value, r["data"]["value"])
                continue

            entry = {
                'id': r["id"],
                'type': r["data"]["type"],
                'name': self._full_name(r["data"]["host"]),
                'ttl': None,
                'options': {}
            }
            
            if r["data"]["type"] in ("CNAME"):
                entry['content'] = r["data"]["value"].rstrip('.')
            else:
                entry['content'] = r["data"]["value"]

            if r["data"]["type"] == "MX":
                entry['options']['mx'] = {
                    'priority': int(r["data"]["opt"])
                }

            entries.append(entry)

        return entries

    def __delete_dns_records_by_id(self, ids):
        if not ids:
            return

        req = []
        for i in ids:
            req.append({
                'del_rec': {
                    'filter': {
                        'id': i
                    }
                }
            })

        self.__plesk_request({
            'dns': req
        })



from __future__ import absolute_import

import json
import logging
import base64
import hmac
import hashlib
import datetime

import requests

from lexicon.providers.base import Provider as BaseProvider


logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['auroradns.eu']

def ProviderParser(subparser):
    subparser.add_argument("--auth-api-key", help="specify API key for authentication")
    subparser.add_argument("--auth-secret-key", help="specify the secret key for authentication")

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.auroradns.eu'

    def authenticate(self):
        zone = None
        payload = self._get('/zones')

        for item in payload:
            if item['name'] == self.domain:
                zone = item

        if not zone:
            raise Exception('No domain found')

        self.domain_id = zone['id']

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, type, name, content):
        data = {'type': type, 'name': self._relative_name(name), 'content': content}
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')
        payload = self._post('/zones/{0}/records'.format(self.domain_id), data)

        logger.debug('create_record: {0}'.format(payload))
        return payload

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        payload = self._get('/zones/{0}/records'.format(self.domain_id))

        # Apply filtering first.
        processed_records = payload
        if type:
            processed_records = [record for record in processed_records if record['type'] == type]
        if name:
            processed_records = [record for record in processed_records if record['name'] == self._relative_name(name)]
        if content:
            processed_records = [record for record in processed_records if record['content'].lower() == content.lower()]

        # Format the records.
        records = []
        for record in processed_records:
            processed_record = {
                'type': record['type'],
                'name': self._full_name(record['name']),
                'ttl': record['ttl'],
                'content': record['content'],
                'id': record['id']
            }
            records.append(processed_record)

        logger.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):
        # Try to find record if no identifier was specified
        if not identifier:
            identifier = self._find_record_identifier(type, name, None)

        data = {}
        if type:
            data['type'] = type
        if name:
            data['name'] = self._relative_name(name)
        if content:
            data['content'] = content
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')

        payload = self._put('/zones/{0}/records/{1}'.format(self.domain_id, identifier), data)

        logger.debug('update_record: %s', payload)
        return payload

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        # Try to find record if no identifier was specified
        delete_record_id = []
        if not identifier:
            records = self.list_records(type, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)
        
        logger.debug('delete_records: %s', delete_record_id)
        
        for record_id in delete_record_id:
            payload = self._delete('/zones/{0}/records/{1}'.format(self.domain_id, record_id))

        logger.debug('delete_record: %s', True)
        return True


    # Helpers
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        t = datetime.datetime.utcnow()
        timestamp = t.strftime('%Y%m%dT%H%M%SZ')
        authorization_header = self._generate_auth_header(action, url, timestamp)

        r = requests.request(action, self.api_endpoint + url, params=query_params,
            data=json.dumps(data),
            headers={
                'X-AuroraDNS-Date': timestamp,
                'Authorization': authorization_header,
                'Content-Type': 'application/json'
            })

        # If the response is a HTTP 409 statusCode, the record already exists: return true.
        if r.status_code == 409:
            return True

        # If the request fails for any other reason, throw an error.
        r.raise_for_status()

        # Try to parse the json, if it not exists, return true.
        try:
            return r.json()
        except:
            return True

    def _generate_auth_header(self, action, url, timestamp):
        secret_key = self._get_provider_option('auth_secret_key')
        api_key = self._get_provider_option('auth_api_key')
        sig = action + url + timestamp

        signature = base64.b64encode(hmac.new(
            secret_key.encode('utf-8'), sig.encode('utf-8'),
            digestmod=hashlib.sha256).digest())

        auth = api_key + ':' + signature.decode('utf-8')
        auth_b64 = base64.b64encode(auth.encode('utf-8'))
        return 'AuroraDNSv1 %s' % (auth_b64.decode('utf-8'))

    def _find_record_identifier(self, type, name, content):
        records = self.list_records(type, name, content)
        logger.debug('records: %s', records)
        if len(records) == 1:
            return records[0]['id']
        else:
            raise Exception('Record identifier could not be found. Try to provide an identifier')



from __future__ import absolute_import

import contextlib
import hmac
import json
import locale
import logging
from email.utils import formatdate
from hashlib import sha1

import requests
from builtins import bytes

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['dnsmadeeasy']

def ProviderParser(subparser):
    subparser.add_argument("--auth-username", help="specify username for authentication")
    subparser.add_argument("--auth-token", help="specify token for authentication")

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = self._get_provider_option('api_endpoint') or 'https://api.dnsmadeeasy.com/V2.0'

    def authenticate(self):

        try:
            payload = self._get('/dns/managed/name', {'domainname': self.domain})
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                payload = {}
            else:
                raise e

        if not payload or not payload['id']:
            raise Exception('No domain found')

        self.domain_id = payload['id']


    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, type, name, content):
        record = {
            'type': type,
            'name': self._relative_name(name),
            'value': content,
            'ttl': self._get_lexicon_option('ttl')
        }
        payload = {}
        try:
            payload = self._post('/dns/managed/{0}/records/'.format(self.domain_id), record)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code != 400:
                raise

                # http 400 is ok here, because the record probably already exists
        logger.debug('create_record: %s', 'name' in payload)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        filter = {}
        if type:
            filter['type'] = type
        if name:
            filter['recordName'] = self._relative_name(name)
        payload = self._get('/dns/managed/{0}/records'.format(self.domain_id), filter)

        records = []
        for record in payload['data']:
            processed_record = {
                'type': record['type'],
                'name': '{0}.{1}'.format(record['name'], self.domain),
                'ttl': record['ttl'],
                'content': record['value'],
                'id': record['id']
            }

            processed_record = self._clean_TXT_record(processed_record)
            records.append(processed_record)

        if content:
            records = [record for record in records if record['content'].lower() == content.lower()]

        logger.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):

        data = {
            'id': identifier,
            'ttl': self._get_lexicon_option('ttl')
        }

        if name:
            data['name'] = self._relative_name(name)
        if content:
            data['value'] = content
        if type:
            data['type'] = type

        payload = self._put('/dns/managed/{0}/records/{1}'.format(self.domain_id, identifier), data)

        logger.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(type, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)
        
        logger.debug('delete_records: %s', delete_record_id)
        
        for record_id in delete_record_id:
            payload = self._delete('/dns/managed/{0}/records/{1}'.format(self.domain_id, record_id))

        # is always True at this point, if a non 200 response is returned an error is raised.
        logger.debug('delete_record: %s', True)
        return True


    # Helpers

    def _request(self, action='GET',  url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'x-dnsme-apiKey': self._get_provider_option('auth_username')
        }
        default_auth = None

        # Date string in HTTP format e.g. Sat, 12 Feb 2011 20:59:04 GMT
        request_date = formatdate(usegmt=True)

        hashed = hmac.new(bytes(self._get_provider_option('auth_token'), 'ascii'),
                          bytes(request_date, 'ascii'), sha1)

        default_headers['x-dnsme-requestDate'] = request_date
        default_headers['x-dnsme-hmac'] = hashed.hexdigest()

        r = requests.request(action, self.api_endpoint + url, params=query_params,
                             data=json.dumps(data),
                             headers=default_headers,
                             auth=default_auth)
        r.raise_for_status()  # if the request fails for any reason, throw an error.

        # PUT and DELETE actions dont return valid json.
        if action == 'DELETE' or action == 'PUT':
            return r.text
        return r.json()



import logging
import os
import re
import yaml
import warnings

LOGGER = logging.getLogger(__name__)

class ConfigResolver(object):
    """
    Highly customizable configuration resolver object, that gets configuration parameters
    from various sources with a precedence order. Sources and their priority are configured 
    by calling the with* methods of this object, in the decreasing priority order.

    A configuration parameter can be retrieved using the resolve() method. The configuration parameter
    key needs to conform to a namespace, whose delimeters is ':'. Two namespaces will be used in 
    the context of Lexicon:
        * the parameters relevant for Lexicon itself: 'lexicon:global_parameter'
        * the parameters specific to a DNS provider: 'lexicon:cloudflare:cloudflare_parameter'

    Example:
        # This will resolve configuration parameters from environment variables, 
        # then from a configuration file named '/my/path/to/lexicon.yml'.
        $ from lexicon.config import Config
        $ config = Config()
        $ config.with_env().with_config_file()
        $ print(config.resolve('lexicon:delegated'))
        $ print(config.resolve('lexicon:cloudflare:auth_token))

    Config can resolve parameters for Lexicon and providers from:
        * environment variables
        * arguments parsed by ArgParse library
        * YAML configuration files, generic or specific to a provider
        * any object implementing the underlying ConfigSource class

    Each parameter will be resolved against each source, and value from the higher priority source
    is returned. If a parameter could not be resolve by any source, then None will be returned.
    """

    def __init__(self):
        super(ConfigResolver, self).__init__()
        self._config_sources = []

    def resolve(self, config_key):
        """
        Resolve the value of the given config parameter key. Key must be correctly scoped for Lexicon, and
        optionally for the DNS provider for which the parameter is consumed. For instance:
            * config.resolve('lexicon:delegated') will get the delegated config parameter for Lexicon
            * config.resolve('lexicon:cloudflare:auth_token') will get 
            the auth_token config parameter consumed by cloudflare DNS provider.

        Value is resolved against each configured source, and value from the highest priority source
        is returned. None will be returned if the given config parameter key could not be resolved
        from any source.
        """
        for config_source in self._config_sources:
            value = config_source.resolve(config_key)
            if value:
                return value

        return None

    def add_config_source(self, config_source, position = None):
        self._config_sources.insert(position if position is not None else len(self._config_sources), config_source)

    def with_config_source(self, config_source):
        """
        Configure current resolver to use the provided ConfigSource instance to be used as a source.
        See documentation of ConfigSource to see how to implement correctly a ConfigSource.
        """
        self.add_config_source(config_source)
        return self

    def with_env(self):
        """
        Configure current resolver to use available environment variables as a source.
        Only environment variables starting with 'LEXICON' or 'LEXICON_[PROVIDER]' 
        will be taken into account.
        """
        return self.with_config_source(EnvironmentConfigSource())

    def with_args(self, argparse_namespace):
        """
        Configure current resolver to use a Namespace object given by a ArgParse instance
        using arg_parse() as a source. This method is typically used to allow a ConfigResolver
        to get parameters from the command line.

        It is assumed that the argument parser have already checked that provided arguments are
        valid for Lexicon or the current provider. No further namespace check on parameter keys will
        be done here. Meaning that if 'lexicon:cloudflare:auth_token' is asked, any auth_token present
        in the given Namespace object will be returned.
        """
        return self.with_config_source(ArgsConfigSource(argparse_namespace))

    def with_dict(self, dict_object):
        """
        Configure current resolver to use the given dict object, scoped to lexicon namespace.
        
        Example of valid dict object for lexicon:
            {
                'delegated': 'onedelegated',
                'cloudflare': {
                    'auth_token': 'SECRET_TOKEN'
                }
            }
            => Will define properties 'lexicon:delegated' and 'lexicon:cloudflare:auth_token'
        """
        return self.with_config_source(DictConfigSource(dict_object))

    def with_config_file(self, file_path):
        """
        Configure current resolver to use a YAML configuration file specified on the given path.
        This file provides configuration parameters for Lexicon and any DNS provider.

        Typical format is:
            $ cat lexicon.yml
            # Will define properties 'lexicon:delegated' and 'lexicon:cloudflare:auth_token'
            delegated: 'onedelegated'
            cloudflare:
            auth_token: SECRET_TOKEN
        """
        return self.with_config_source(FileConfigSource(file_path))

    def with_provider_config_file(self, provider_name, file_path):
        """
        Configure current resolver to use a YAML configuration file specified on the given path.
        This file provides configuration parameters for a DNS provider exclusively.

        Typical format is:
            $ cat lexicon_cloudflare.yml
            # Will define properties 'lexicon:cloudflare:auth_token' and 'lexicon:cloudflare:auth_username'
            auth_token: SECRET_TOKEN
            auth_username: USERNAME

        NB: If file_path is not specified, '/etc/lexicon/lexicon_[provider].yml' will be taken
        by default, with [provider] equals to the given provider_name parameter.
        """
        return self.with_config_source(ProviderFileConfigSource(provider_name, file_path))

    def with_config_dir(self, dir_path):
        """
        Configure current resolver to use every valid YAML configuration files available in the
        given directory path. To be taken into account, a configuration file must conform to the
        following naming convention:
            * 'lexicon.yml' for a global Lexicon config file (see with_config_file doc)
            * 'lexicon_[provider].yml' for a DNS provider specific configuration file, with
            [provider] equals to the DNS provider name (see with_provider_config_file doc)

        Example:
            $ ls /etc/lexicon
            lexicon.yml # global Lexicon configuration file
            lexicon_cloudflare.yml # specific configuration file for clouflare DNS provder
        """
        lexicon_provider_config_files = []
        lexicon_config_files = []

        for path in os.listdir(dir_path):
            path = os.path.join(dir_path, path)
            if os.path.isfile(path):
                basename = os.path.basename(path)
                search = re.search(r'^lexicon(?:_(\w+)|)\.yml$', basename)
                if search:
                    provider = search.group(1)
                    if provider:
                        lexicon_provider_config_files.append((provider, path))
                    else:
                        lexicon_config_files.append(path)

        for lexicon_provider_config_file in lexicon_provider_config_files:
            self.with_provider_config_file(lexicon_provider_config_file[0], 
                                           lexicon_provider_config_file[1])

        for lexicon_config_file in lexicon_config_files:
            self.with_config_file(lexicon_config_file)

        return self

    def with_legacy_dict(self, legacy_dict_object):
        warnings.warn(DeprecationWarning('Legacy configuration object has been used '
                                         'to load the ConfigResolver.'))
        return self.with_config_source(LegacyDictConfigSource(legacy_dict_object))

class ConfigSource(object):
    """
    Base class to implement a configuration source for a ConfigResolver.
    The relevant method to override is resolve(self, config_parameter).
    """

    def resolve(self, config_parameter):
        """
        Using the given config_parameter value (in the form of 'lexicon:config_key' or 
        'lexicon:[provider]:config_key'), try to get the associated value.

        None must be returned if no value could be found.

        Must be implemented by each ConfigSource concrete child class.
        """
        raise NotImplementedError('The method resolve(config_parameter) '
                                  'must be implemented in the concret sub-classes.')

class EnvironmentConfigSource(ConfigSource):

    def __init__(self):
        super(EnvironmentConfigSource, self).__init__()
        self._parameters = {}
        for (key, value) in os.environ.items():
            if key.startswith('LEXICON_'):
                self._parameters[key] = value

    def resolve(self, config_parameter):
        # First try, with a direct conversion of the config_parameter: 
        #   * lexicon:provider:auth_my_config => LEXICON_PROVIDER_AUTH_MY_CONFIG
        #   * lexicon:provider:my_other_config => LEXICON_PROVIDER_AUTH_MY_OTHER_CONFIG
        #   * lexicon:my_global_config => LEXICON_MY_GLOBAL_CONFIG
        environment_variable = re.sub(':', '_', config_parameter).upper()
        value = self._parameters.get(environment_variable, None)
        if value:
            return value

        # Second try, with the legacy naming convention for specific provider config: 
        #   * lexicon:provider:auth_my_config => LEXICON_PROVIDER_MY_CONFIG
        # Users get a warning about this deprecated usage.
        environment_variable_legacy = re.sub(r'(.*)_AUTH_(.*)', r'\1_\2',
                                             environment_variable).upper()
        value = self._parameters.get(environment_variable_legacy, None)
        if value:
            LOGGER.warning(('Warning: Use of environment variable %s is deprecated. '
                            'Try %s instead.'),
                           environment_variable_legacy,
                           environment_variable)
            return value

        return None

class ArgsConfigSource(ConfigSource):

    def __init__(self, namespace):
        super(ArgsConfigSource, self).__init__()
        self._parameters = vars(namespace)

    def resolve(self, config_key):
        # We assume here that the namespace provided has already done its job,
        # by validating that all given parameters are relevant for Lexicon or the current provider.
        # So we ignore the namespaces 'lexicon:' and 'lexicon:provider' in given config key.
        splitted_config_key = config_key.split(':')

        return self._parameters.get(splitted_config_key[-1], None)

class DictConfigSource(ConfigSource):

    def __init__(self, dict_object):
        super(DictConfigSource, self).__init__()
        self._parameters = dict_object

    def resolve(self, config_key):
        splitted_config_key = config_key.split(':')
        # Note that we ignore 'lexicon:' in the iteration,
        # as the dict object is already scoped to lexicon.
        cursor = self._parameters
        for current in splitted_config_key[1:-1]:
            cursor = cursor.get(current, {})

        return cursor.get(splitted_config_key[-1], None)

class FileConfigSource(DictConfigSource):

    def __init__(self, file_path):
        with open(file_path, 'r') as stream:
            yaml_object = yaml.load(stream) or {}

        super(FileConfigSource, self).__init__(yaml_object)

class ProviderFileConfigSource(FileConfigSource):

    def __init__(self, provider_name, file_path):
        super(ProviderFileConfigSource, self).__init__(file_path)
        # Scope the loaded config file into provider namespace
        self._parameters = {provider_name: self._parameters}

class LegacyDictConfigSource(DictConfigSource):

    def __init__(self, dict_object):
        provider_name = dict_object.get('provider_name')
        if not provider_name:
            raise AttributeError('Error, key provider_name is not defined.'
                                 'LegacyDictConfigSource cannot scope correctly '
                                 'the provider specific options.')

        generic_parameters = [
            'domain', 'action', 'provider_name', 'delegated', 'identifier', 'type' , 'name',
            'content', 'ttl', 'priority', 'log_level', 'output']

        provider_options = {}
        refactor_dict_object = {}
        refactor_dict_object[provider_name] = provider_options

        for (key, value) in dict_object.items():
            if key not in generic_parameters:
                provider_options[key] = value
            else:
                refactor_dict_object[key] = value

        super(LegacyDictConfigSource, self).__init__(refactor_dict_object)

def non_interactive_config_resolver():
    """
    Create a typical config resolver in a non-interactive context (eg. lexicon used as a library).
    Configuration will be resolved againts env variables and lexicon config files in working dir.
    """
    return ConfigResolver().with_env().with_config_dir(os.getcwd())

def legacy_config_resolver(legacy_dict):
    """
    With the old legacy approach, we juste got a plain configuration dict object.
    Custom logic was to enrich this configuration with env variables.

    This function create a resolve that respect the expected behavior, by using the relevant
    ConfigSources, and we add the config files from working directory.
    """
    return ConfigResolver().with_legacy_dict(legacy_dict).with_env().with_config_dir(os.getcwd())


from __future__ import absolute_import

import logging

from lexicon.providers.base import Provider as BaseProvider

try:
    import SoftLayer
except ImportError:
    pass

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['softlayer.com']

def ProviderParser(subparser):
    subparser.add_argument("--auth-username", help="specify username for authentication")
    subparser.add_argument("--auth-api-key", help="specify API private key for authentication")

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None

        username = self._get_provider_option('auth_username')
        api_key = self._get_provider_option('auth_api_key')

        if not username or not api_key:
            raise Exception("No username and/or api key was specified")

        sl_client = SoftLayer.create_client_from_env(username=username, api_key=api_key)
        self.sl_dns = SoftLayer.managers.dns.DNSManager(sl_client)


    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider, so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason, of if the domain does not exist.
    def authenticate(self):
        domain = self.domain

        payload = self.sl_dns.resolve_ids(domain)

        if len(payload) < 1:
            raise Exception('No domain found')
        if len(payload) > 1:
            raise Exception('Too many domains found. This should not happen')

        logger.debug('domain id: %s', payload[0])
        self.domain_id = payload[0]


    # Create record. If record already exists with the same content, do nothing
    def create_record(self, type, name, content):
        records = self.list_records(type,name,content)
        if len(records) > 0:
            # Nothing to do, record already exists
            logger.debug('create_record: already exists')
            return True

        name = self._relative_name(name)
        ttl = self._get_lexicon_option('ttl')
        payload = self.sl_dns.create_record(self.domain_id,name,type,content,ttl)

        logger.debug('create_record: %s', payload)
        return True


    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        ttl=None
        if name:
            name = self._relative_name(name)

        payload = self.sl_dns.get_records(self.domain_id,ttl,content,name,type)

        records = []
        for record in payload:
            processed_record = {
                'type': record['type'].upper(),
                'name': self._full_name(record['host']),
                'ttl': record['ttl'],
                'content': record['data'],
                'id': record['id']
            }
            records.append(processed_record)

        logger.debug('list_records: %s', records)
        return records


    # Update a record.
    # If an identifier is specified, use it, otherwise do a lookup using type and name.
    def update_record(self, identifier=None, type=None, name=None, content=None):
        if not identifier:
            records = self.list_records(type, name)
            if len(records) == 1:
                identifier = records[0]['id']
            else:
                raise Exception('Record identifier could not be found.')

        record = { 'id': identifier }
        if type:
            record['type'] = type
        if name:
            record['host'] = self._relative_name(name)
        if content:
            record['data'] = content
        if self._get_lexicon_option('ttl'):
            record['ttl'] = self._get_lexicon_option('ttl')

        payload = self.sl_dns.edit_record(record)

        logger.debug('update_record: %s', payload)
        return True


    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(type, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)
        
        logger.debug('delete_records: %s', delete_record_id)
        
        for record_id in delete_record_id:
            payload = self.sl_dns.delete_record(record_id)

        logger.debug('delete_record: %s', True)
        return True




from __future__ import absolute_import

import logging
import requests
import json
import hashlib

from lexicon.providers.base import Provider as BaseProvider
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['godaddy.com']

def ProviderParser(subparser):
    subparser.add_argument('--auth-key', help='specify the key to access the API')
    subparser.add_argument('--auth-secret', help='specify the secret to access the API')

# Implements the DNS GoDaddy provider.
# Some general remarks about this provider, because it uses a weirdly designed API.
# Indeed, there is no direct way to insert, update or delete a specific record. Furthermore, there is no unique identifier for a record.
# Instead GoDaddy use a replace approach: for a given set of records one can replace this set with a new set sent throught API.
# For the sake of simplicity and consistency accoss the provider edit methods, the set will be always all records in the DNS zone.
# With this approach:
#   - adding a record consists in appending a record to the obtained set and call replace with the updated set,
#   - updating a record consists in modifying a record in the obtained set and call replace with the updated set,
#   - deleting a record consists in removing a record in the obtained set and call replace with the updated set.
# In parallel, as said before, there is no unique identifier. This provider then implement a pseudo-identifier,
#   to allow an easy update or delete using the '--identifier' lexicon parameter.
# But you need to call the 'list' command just before executing and update/delete action, because identifier value
#   is tied to the content of the record, and will change anytime something is changed in the record.
class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.godaddy.com/v1'

    def authenticate(self):
        domain = self.domain

        result = self._get('/domains/{0}'.format(domain))
        self.domain_id = result['domainId']

    def list_records(self, type=None, name=None, content=None):
        domain = self.domain

        url = '/domains/{0}/records'.format(domain)
        if type:
            url += '/{0}'.format(type)
        if name:
            url += '/{0}'.format(self._relative_name(name))

        raws = self._get(url)

        records = []
        for raw in raws:
            records.append({
                'id': Provider._identifier(raw),
                'type': raw['type'],
                'name': self._full_name(raw['name']),
                'ttl': raw['ttl'],
                'content': raw['data']
            })

        if content:
            records = [record for record in records if record['data'] == content]

        LOGGER.debug('list_records: %s', records)

        return records

    def create_record(self, type, name, content):
        domain = self.domain
        relative_name = self._relative_name(name)
        ttl = self._get_lexicon_option('ttl')

        # Retrieve existing data in DNS zone.
        records = self._get('/domains/{0}/records'.format(domain))

        # Check if a record already matches given parameters
        for record in records:
            if record['type'] == type and self._relative_name(record['name']) == relative_name and record['data'] == content:
                LOGGER.debug('create_record (ignored, duplicate): %s %s %s', type, name, content)
                return True

        # Append a new entry corresponding to given parameters.
        data = {'type': type, 'name': relative_name, 'data': content}
        if ttl:
            data['ttl'] = ttl

        records.append(data)

        # Synchronize data with inserted record into DNS zone.
        self._put('/domains/{0}/records'.format(domain), records)

        LOGGER.debug('create_record: %s %s %s', type, name, content)

        return True

    def update_record(self, identifier, type=None, name=None, content=None):
        # No identifier is used with GoDaddy. 
        # We can rely either:
        #   - only on type/name to get the relevant records, both of them are required or we will could update to much records ...,
        #   - or by the pseudo-identifier provided
        # Furthermore for type/name approach, we cannot update all matching records, as it would lead 
        #   to an error (two entries of same type + name cannot have the same content).
        # So for type/name approach, we search first matching record for type/name on which content is different, 
        #   and we update it before synchronizing the DNS zone.
        if not identifier and not type:
            raise Exception('ERROR: type is required')
        if not identifier and not name:
            raise Exception('ERROR: name is required')

        domain = self.domain
        relative_name = None
        if name:
            relative_name = self._relative_name(name)

        # Retrieve existing data in DNS zone.
        records = self._get('/domains/{0}/records'.format(domain))

        # Get the record to update: 
        #   - either explicitly by its identifier, 
        #   - or the first matching by its type+name where content does not match (first match, see first method comment for explanation).
        for record in records:
            if ((identifier and Provider._identifier(record) == identifier) or
                (not identifier and record['type'] == type and self._relative_name(record['name']) == relative_name and record['data'] != content)):
                record['data'] = content
                break
        
        # Synchronize data with updated records into DNS zone.
        self._put('/domains/{0}/records'.format(domain), records)

        LOGGER.debug('update_record: %s %s %s', type, name, content)

        return True

    def delete_record(self, identifier=None, type=None, name=None, content=None):
        # For the LOL. GoDaddy does not accept an empty array when updating a particular set of records.
        # It means that you cannot request to remove all records matching a particular type and/or name.
        # Instead, we get ALL records in the DNS zone, update the set, and replace EVERYTHING in the DNS zone.
        # You will always have at minimal NS/SRV entries in the array, otherwise your DNS zone is broken,
        #   and updating the zone is the least of your problem ...
        domain = self.domain

        # Retrieve all records in the DNS zone
        records = self._get('/domains/{0}/records'.format(domain))

        relative_name = None
        if name:
            relative_name = self._relative_name(name)

        # Filter out all records which matches the pattern (either identifier, or some combination of type/name/content).
        filtered_records = []
        if identifier:
            filtered_records = [record for record in records if Provider._identifier(record) != identifier]
        else:
            for record in records:
                if ((not type and not relative_name and not content) or
                    (type and not relative_name and not content and record['type'] != type) or
                    (not type and relative_name and not content and self._relative_name(record['name']) != relative_name) or
                    (not type and not relative_name and content and record['data'] != content) or
                    (type and relative_name and not content and (record['type'] != type or self._relative_name(record['name']) != relative_name)) or
                    (type and not relative_name and content and (record['type'] != type or record['data'] != content)) or
                    (not type and relative_name and content and (self._relative_name(record['name']) != relative_name or record['data'] != content)) or
                    (type and relative_name and content and (record['type'] != type or self._relative_name(record['name']) != relative_name or record['data'] != content))):
                    filtered_records.append(record)

        # Synchronize data with expurged entries into DNS zone.
        self._put('/domains/{0}/records'.format(domain), filtered_records)

        LOGGER.debug('delete_records: %s %s %s', type, name, content)

        return True

    # GoDaddy provides no identifier for a record, which is a problem where identifiers can be used (delete and update).
    # To circumvent this, we implement a pseudo-identifier, which is basically a hash of type+name+content of a given record.
    # It is far from perfect, as the identifier will change each time we change something in the record ...
    # But at least, one can use 'lexicon godaddy list ...' then 'lexicon godaddy update --identifier ...' to modify specific record.
    # However, 'lexicon godaddy list ...' should be called each time DNS zone had been changed to calculate new identifiers.
    @staticmethod
    def _identifier(record):
        sha256 = hashlib.sha256()
        sha256.update(('type=' + record.get('type', '') + ',').encode('utf-8'))
        sha256.update(('name=' + record.get('name', '') + ',').encode('utf-8'))
        sha256.update(('data=' + record.get('data', '') + ',').encode('utf-8'))
        return sha256.hexdigest()[0:7]

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if not data:
            data = {}
        if not query_params:
            query_params = {}

        # When editing DNS zone, API is unavailable for few seconds (until modifications are propagated)
        # In this case, call to API will return 409 HTTP error.
        # We use the Retry extension to retry the requests until we get a processable reponse (402 HTTP status, or an HTTP error != 409)
        retries = Retry(
            total=10,
            backoff_factor=0.5,
            status_forcelist=[409],
            method_whitelist=frozenset(['GET', 'PUT', 'POST', 'DELETE', 'PATCH'])
        )

        session = requests.Session()
        session.mount('https://', HTTPAdapter(max_retries=retries))

        result = session.request(action, self.api_endpoint + url,
                                  params=query_params,
                                  data=json.dumps(data),
                                  headers={
                                      'Content-Type': 'application/json',
                                      'Accept': 'application/json',
                                      # GoDaddy use a key/secret pair to authenticate
                                      'Authorization': 'sso-key {0}:{1}'.format(
                                          self._get_provider_option('auth_key'),
                                          self._get_provider_option('auth_secret'))
                                  })

        result.raise_for_status()

        try:
            # Return the JSON body response if exists.
            return result.json()
        except ValueError:
            # For some requests command (eg. PUT), GoDaddy will not return any JSON, just an HTTP status without body.
            return None



import argparse
import importlib
import pkgutil
import os

import pkg_resources

from lexicon import providers as providers_package

def generate_base_provider_parser():
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('action', help='specify the action to take', default='list', choices=['create', 'list', 'update', 'delete'])
    parser.add_argument('domain', help='specify the domain, supports subdomains as well')
    parser.add_argument('type', help='specify the entry type', default='TXT', choices=['A', 'AAAA', 'CNAME', 'MX', 'NS', 'SOA', 'TXT', 'SRV', 'LOC'])

    parser.add_argument('--name', help='specify the record name')
    parser.add_argument('--content', help='specify the record content')
    parser.add_argument('--ttl', type=int, help='specify the record time-to-live')
    parser.add_argument('--priority', help='specify the record priority')
    parser.add_argument('--identifier', help='specify the record for update or delete actions')
    parser.add_argument('--log_level', help='specify the log level', default='ERROR', choices=['CRITICAL','ERROR','WARNING','INFO','DEBUG','NOTSET'])
    parser.add_argument('--output', 
                        help='specify the type of output: by default a formatted table (TABLE), a formatted table without header (TABLE-NO-HEADER), a JSON string (JSON) or no output (QUIET)',
                        default='TABLE', choices=['TABLE', 'TABLE-NO-HEADER', 'JSON', 'QUIET'])
    return parser

def generate_cli_main_parser():
    providers = []
    for _, modname, _ in pkgutil.iter_modules(providers_package.__path__):
        if modname != 'base':
            providers.append(modname)
    providers = sorted(providers)

    parser = argparse.ArgumentParser(description='Create, Update, Delete, List DNS entries')
    try:
        version = pkg_resources.get_distribution('dns-lexicon').version
    except pkg_resources.DistributionNotFound:
        version = 'unknown'
    parser.add_argument('--version', help='show the current version of lexicon', action='version', version='%(prog)s {0}'.format(version))
    parser.add_argument('--delegated', help='specify the delegated domain')
    subparsers = parser.add_subparsers(dest='provider_name', help='specify the DNS provider to use')
    subparsers.required = True

    for provider in providers:
        provider_module = importlib.import_module('lexicon.providers.' + provider)
        provider_parser = getattr(provider_module, 'ProviderParser')

        subparser = subparsers.add_parser(provider, help='{0} provider'.format(provider), parents=[generate_base_provider_parser()])
        provider_parser(subparser)

    return parser



from __future__ import absolute_import

import json
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['memset.com']

def ProviderParser(subparser):
    subparser.add_argument("--auth-token", help="specify API key for authentication")


class Provider(BaseProvider):
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.memset.com/v1/json'

    def authenticate(self):
        payload = self._get('/dns.zone_domain_info', {
            'domain': self.domain
        })
        if not payload['zone_id']:
            raise Exception('No domain found')
        self.domain_id = payload['zone_id']

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, type, name, content):
        data = {'type': type, 'record': self._relative_name(name), 'address': content}
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')
        data['zone_id'] = self.domain_id
        check_exists = self.list_records(type=type, name=name, content=content)
        if not len(check_exists) > 0:
            payload = self._get('/dns.zone_record_create', data)
            if payload['id']:
                self._get('/dns.reload')
                logger.debug('create_record: %s', payload['id'])
                return payload['id']
        else:
            return check_exists

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        payload = self._get('/dns.zone_info', {
            'id': self.domain_id
            })
        records = []
        for record in payload['records']:
            processed_record = {
                'type': record['type'],
                'name': self._full_name(record['record']),
                'ttl': record['ttl'],
                'content': record['address'],
                'id': record['id']
            }
            if name:
                name = self._full_name(name)
            if (processed_record['type'] == type):
                if (name is not None and content is not None):
                    if processed_record['name'] == name and processed_record['content'] == content:
                        records.append(processed_record)
                elif (name is not None and content is None):
                    if processed_record['name'] == name:
                        records.append(processed_record)
                elif (name is None and content is not None):
                    if processed_record['content'] == content:
                        records.append(processed_record)
                else:
                    records.append(processed_record)

        logger.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):
        data = {}
        if not identifier:
            records = self.list_records(type, self._relative_name(name))
            if len(records) == 1:
                identifier = records[0]['id']
            else:
                raise Exception('Record identifier could not be found.')
        if type:
            data['type'] = type
        if name:
            data['record'] = self._relative_name(name)
        if content:
            data['address'] = content
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')
        data['id'] = identifier
        data['zone_id'] = self.domain_id

        payload = self._get('/dns.zone_record_update', data)
        if payload['id']:
            self._get('/dns.reload')
            logger.debug('update_record: %s', payload['id'])
            return payload['id']

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(type, self._relative_name(name), content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)
        
        logger.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            payload = self._get('/dns.zone_record_delete', {'id': record_id})

        if len(record_id) > 0:
            self._get('/dns.reload')

        logger.debug('delete_record: %s', True)
        return True

    # Helpers
    def _request(self, action='GET',  url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        r = requests.request(action, self.api_endpoint + url, params=query_params,
                             data=json.dumps(data),
                             auth=(self._get_provider_option('auth_token'), 'x'),
                             headers={'Content-Type': 'application/json'})
        r.raise_for_status()  # if the request fails for any reason, throw an error.
        return r.json()




from __future__ import absolute_import

import re, logging
from sys import stderr
from os import environ
from time import sleep
from requests import Session
# Due to optional requirement
try:
    from bs4 import BeautifulSoup
except ImportError:
   pass

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['he.net']

def ProviderParser(subparser):
    subparser.description = """A provider for Hurricane Electric DNS.
        NOTE: THIS DOES NOT WORK WITH 2-FACTOR AUTHENTICATION.
              YOU MUST DISABLE IT IF YOU'D LIKE TO USE THIS PROVIDER.
        """
    subparser.add_argument(
        '--auth-username',
        help='specify username for authentication'
    )
    subparser.add_argument(
        '--auth-password',
        help='specify password for authentication',
    )


class Provider(BaseProvider):
    """
        he.net provider
    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain = self.domain
        self.domain_id = None

    def authenticate(self):
        """
        """
        # Create the session GET the login page to retrieve a session cookie
        self.session = Session()    
        self.session.get(
            "https://dns.he.net/"
        )

        # Hit the login page with authentication info to login the session
        login_response = self.session.post(
           "https://dns.he.net",
            data={
                "email": self._get_provider_option('auth_username') or '',
                "pass": self._get_provider_option('auth_password') or ''
            }
        )

        # Parse in the HTML, if the div containing the error message is found, error
        html = BeautifulSoup(login_response.content, "html.parser")
        if html.find("div", {"id": "dns_err"}) is not None:
            logger.warning("HE login failed, check HE_USER and HE_PASS")
            return False

        # Make an authenticated GET to the DNS management page
        zones_response = self.session.get(
            "https://dns.he.net"
        )

        html = BeautifulSoup(zones_response.content, "html.parser")
        zone_img = html.find("img", {"name": self.domain, "alt": "delete"})

        # If the tag couldn't be found, error, otherwise, return the value of the tag
        if zone_img is None:
            logger.warning("Domain {0} not found in account".format(self.domain))
            raise AssertionError("Domain {0} not found in account".format(self.domain))

        self.domain_id = zone_img["value"]
        logger.debug("HENET domain ID: {}".format(self.domain_id))
        return True

    # Create record. If record already exists with the same content, do nothing
    def create_record(self, type, name, content):
        logger.debug("Creating record for zone {0}".format(name))
        # Pull a list of records and check for ours
        records = self.list_records(type=type, name=name, content=content)
        if len(records) >= 1:
            logger.warning("Duplicate record {} {} {}, NOOP".format(type, name, content))
            return True
        data={
            "account": "",
            "menu": "edit_zone",
            "Type": type,
            "hosted_dns_zoneid": self.domain_id,
            "hosted_dns_recordid": "",
            "hosted_dns_editzone": "1",
            "Priority": "",
            "Name": name,
            "Content": content,
            "TTL": "3600",
            "hosted_dns_editrecord": "Submit"
        }
        ttl = self._get_lexicon_option('ttl')
        if ttl:
            if ttl <= 0: 
                data['TTL'] = "3600"
            else:
                data['TTL'] = str(ttl)
        prio = self._get_lexicon_option('priority')
        if prio:
            if prio <= 0: 
                data['Priority'] = "10"
            else:
                data['Priority'] = str(prio)
        create_response = self.session.post(
            "https://dns.he.net/index.cgi", data=data
        )
        # Pull a list of records and check for ours
        records = self.list_records(name=name)
        if len(records) >= 1:
            logger.info("Successfully added record {}".format(name))
            return True
        else:
            logger.info("Failed to add record {}".format(name))
            return False

    # List all records. Return an empty list if no records found.
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is
    # received.
    def list_records(self, type=None, name=None, content=None, id=None):
        records = []
        # Make an authenticated GET to the DNS management page
        edit_response = self.session.get(
            "https://dns.he.net/?hosted_dns_zoneid={0}&menu=edit_zone&hosted_dns_editzone".format(self.domain_id)
        )

        # Parse the HTML response, and list the table rows for DNS records
        html = BeautifulSoup(edit_response.content, "html.parser")
        def is_dns_tr_type(klass):
            return klass and re.compile("dns_tr").search(klass)
        records = html.findAll("tr", class_=is_dns_tr_type)

        # If the tag couldn't be found, error, otherwise, return the value of the tag
        if records is None or len(records) == 0:
            logger.warning("Domains not found in account")
        else:
            new_records = []
            for dns_tr in records:
                tds = dns_tr.findAll("td")
                # Process HTML in the TR children to derive each object
                rec = {}
                rec['zone_id'] = tds[0].string
                rec['id'] = tds[1].string
                rec['name'] = tds[2].string
                # the 4th entry is a comment
                type_elem = tds[3].find("span", class_='rrlabel')
                if type_elem:
                    rec['type'] = type_elem.string
                else:
                    rec['type'] = None
                rec['ttl'] = tds[4].string
                if tds[5].string != '-':
                    rec['priority'] = tds[5]
                rec['content'] = tds[6].string
                if tds[7].string == '1':
                    rec['is_dynamic'] = True
                else:
                    rec['is_dynamic'] = False
                rec = self._clean_TXT_record(rec)
                new_records.append(rec)
            records = new_records
            if id:
                logger.debug("Filtering {} records by id: {}".format(len(records), id))
                records = [record for record in records if record['id'] == id]
            if type:
                logger.debug("Filtering {} records by type: {}".format(len(records), type))
                records = [record for record in records if record['type'] == type]
            if name:
                logger.debug("Filtering {} records by name: {}".format(len(records), name))
                if name.endswith('.'):
                    name = name[:-1]
                records = [record for record in records if name in record['name'] ]
            if content:
                logger.debug("Filtering {} records by content: {}".format(len(records), content.lower()))
                records = [record for record in records if record['content'].lower() == content.lower()]
            logger.debug("Final records ({}): {}".format(len(records), records))
        return records

    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):
        # Delete record if it exists
        self.delete_record(identifier, type, name, content)
        return self.create_record(type, name, content)

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_record_ids = []
        if not identifier:
            records = self.list_records(type, name, content)
            delete_record_ids = [record['id'] for record in records]
        else:
            delete_record_ids.append(identifier)
        logger.debug("Record IDs to delete: {}".format(delete_record_ids))
        for rec_id in delete_record_ids:
            # POST to the DNS management UI with form values to delete the record
            delete_response = self.session.post(
                "https://dns.he.net/index.cgi",
                data={
                    "menu": "edit_zone",
                    "hosted_dns_zoneid": self.domain_id,
                    "hosted_dns_recordid": rec_id,
                    "hosted_dns_editzone": "1",
                    "hosted_dns_delrecord": "1",
                    "hosted_dns_delconfirm": "delete"
                }
            )

            # Parse the HTML response, if the <div> tag indicating success isn't found, error
            html = BeautifulSoup(delete_response.content, "html.parser")
            if html.find("div", {"id": "dns_status"}) is None:
                logger.warning("Unable to delete record {}".format(rec_id))
                return False
        return True



from __future__ import absolute_import

import json

import requests

from lexicon.providers.base import Provider as BaseProvider

NAMESERVER_DOMAINS = ['glesys.com']

def ProviderParser(subparser):
    subparser.add_argument("--auth-username", help="specify username (CL12345)")
    subparser.add_argument("--auth-token", help="specify API key")

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.glesys.com'

    def authenticate(self):
        payload = self._get('/domain/list')
        domains = payload['response']['domains']
        for record in domains:
            if record['domainname'] == self.domain:
                # Domain records do not have any id.
                # Since domain_id cannot be None, use domain name as id instead.
                self.domain_id = record['domainname']
                break

        if self.domain_id == None:
            raise Exception('No domain found')

    # Create record. If record already exists with the same content, do nothing.
    def create_record(self, type, name, content):
        existing = self.list_records(type, name, content)
        if len(existing) > 0:
            # Already exists, do nothing.
            return True

        request_data = {
            'domainname': self.domain,
            'host': self._full_name(name),
            'type': type,
            'data': content
        }
        self._addttl(request_data)

        self._post('/domain/addrecord', data=request_data)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        request_data = {
            'domainname': self.domain
        }
        payload = self._post('/domain/listrecords', data=request_data)

        # Convert from Glesys record structure to Lexicon structure.
        processed_records = [self._glesysrecord2lexiconrecord(r) for r in payload['response']['records']]

        if type:
            processed_records = [record for record in processed_records if record['type'] == type]
        if name:
            processed_records = [record for record in processed_records if record['name'] == self._full_name(name)]
        if content:
            processed_records = [record for record in processed_records if record['content'].lower() == content.lower()]

        return processed_records

    # Update a record. Identifier must be specified.
    def update_record(self, identifier, type=None, name=None, content=None):
        request_data = {'recordid': identifier}
        if name:
            request_data['host'] = name
        if type:
            request_data['type'] = type
        if content:
            request_data['data'] = content

        self._addttl(request_data)
        self._post('/domain/updaterecord', data=request_data)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(type, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        for record_id in delete_record_id:
            request_data = {'recordid': record_id}
            self._post('/domain/deleterecord', data=request_data)

        return True

    # Helpers.
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}

        query_params['format'] = 'json'
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }

        credentials = (self._get_provider_option('auth_username'), self._get_provider_option('auth_token'))
        response = requests.request(action,
                                    self.api_endpoint + url,
                                    params=query_params,
                                    data=json.dumps(data),
                                    headers=default_headers,
                                    auth=credentials)

        # if the request fails for any reason, throw an error.
        response.raise_for_status()
        return response.json()

    # Adds TTL parameter if passed as argument to lexicon.
    def _addttl(self, request_data):
        if self._get_lexicon_option('ttl'):
            request_data['ttl'] = self._get_lexicon_option('ttl')

    # From Glesys record structure: [u'domainname', u'recordid', u'type', u'host', u'ttl', u'data']
    def _glesysrecord2lexiconrecord(self, glesys_record):
        return {
            'id': glesys_record['recordid'],
            'type': glesys_record['type'],
            'name': glesys_record['host'],
            'ttl': glesys_record['ttl'],
            'content': glesys_record['data']
        }



from __future__ import absolute_import

import json
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['linode.com']

def ProviderParser(subparser):
    subparser.add_argument("--auth-token", help="specify api key for authentication")

class Provider(BaseProvider):
    
    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.linode.com/v4/'

    def authenticate(self):
        self.domain_id = None
        payload = self._get('domains', query_params={
            'filter': {
                'domain': self.domain
            }
        })
        if len(payload['data']) > 0:
            self.domain_id = payload['data'][0]['id']
        if self.domain_id == None:
            raise Exception('Domain not found')

    def create_record(self, type, name, content):
        if len(self.list_records(type, name, content)) == 0:
            if name:
                name = self._relative_name(name)
            
            self._post('domains/{0}/records'.format(self.domain_id), data={
                'name': name,
                'type': type,
                'target': content,
                'ttl_sec': 0
            })

        return True
    
    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        resources_url = "domains/{0}/records".format(self.domain_id)
        
        if name:
            name = self._relative_name(name)
        
        processed_records = []
        
        payload = self._get(resources_url)
        for page in range(1, payload['pages'] + 1, 1):
            if page > 1:
                payload = self._get(resources_url, query_params={
                    'page': page
                })
            
            resource_list = payload['data']
            if type:
                resource_list = [resource for resource in resource_list if resource['type'] == type]
            if name:
                resource_list = [resource for resource in resource_list if self._relative_name(resource['name']) == name]
            if content:
                resource_list = [resource for resource in resource_list if resource['target'] == content]
            
            for resource in resource_list:
                processed_records.append({
                    'id': resource['id'],
                    'type': resource['type'],
                    'name': self._full_name(resource['name']),
                    'ttl': resource['ttl_sec'],
                    'content': resource['target']
                })
        
        logger.debug('list_records: %s', processed_records)
        return processed_records
    
    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):
        if not identifier:
            resources = self.list_records(type, name, None)
            identifier = resources[0]['id'] if len(resources) > 0 else None
        
        logger.debug('update_record: %s', identifier)
        
        if name:
            name = self._relative_name(name)
        
        url = 'domains/{0}/records/{1}'.format(self.domain_id, identifier)
        self._put(url, data={
            'name': name.lower() if name else None,
            'type': type if type else None,
            'target': content if content else None
        })
        
        return True
    
    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_resource_id = []
        if not identifier:
            resources = self.list_records(type, name, content)
            delete_resource_id = [resource['id'] for resource in resources]
        else:
            delete_resource_id.append(identifier)
        
        logger.debug('delete_records: %s', delete_resource_id)
        
        for resource_id in delete_resource_id:
            self._delete('domains/{0}/records/{1}'.format(
                self.domain_id,
                resource_id
            ))
        
        return True

    # Helpers
    def _request(self, action='GET',  url='', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
            'Authorization': 'Bearer {0}'.format(self._get_provider_option('auth_token'))
        }
        
        request_filter = query_params['filter'] if 'filter' in query_params else None
        if request_filter is not None:
            default_headers['X-Filter'] = json.dumps(request_filter)
            del query_params['filter']
        
        request_url = "{0}{1}".format(self.api_endpoint, url)
        
        r = requests.request(action, request_url, params=query_params,
                             data=json.dumps(data),
                             headers=default_headers)
        r.raise_for_status()  # if the request fails for any reason, throw an error.
        if action == 'DELETE':
            return ''
        else:
            result = r.json()
            return result




#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import print_function

import argparse
import importlib
import logging
import os
import sys
import json

import pkg_resources

from lexicon.client import Client
from lexicon.config import ConfigResolver
from lexicon.parser import generate_cli_main_parser

#based off https://docs.python.org/2/howto/argparse.html

logger = logging.getLogger(__name__)

# Convert returned JSON into a nice table for command line usage
def generate_table_result(logger, output=None, without_header=None):
    try:
        _ = (entry for entry in output)
    except:
        logger.debug('Command output is not iterable, and then cannot be printed with --quiet parameter not enabled.')
        return None

    array = [[
        row.get('id', ''), 
        row.get('type', ''), 
        row.get('name', ''), 
        row.get('content', ''), 
        row.get('ttl', '')] for row in output]

    # Insert header (insert before calculating the max width of each column to take headers size into account)
    if not without_header:
        headers = ['ID', 'TYPE', 'NAME', 'CONTENT', 'TTL']
        array.insert(0, headers)

    columnWidths = [0, 0, 0, 0, 0]
    # Find max width for each column
    for row in array:
        for idx, col in enumerate(row):
            width = len(str(col))
            if width > columnWidths[idx]:
                columnWidths[idx] = width

    # Add a 'nice' separator
    if not without_header:
        array.insert(1, ['-' * columnWidths[idx] for idx in range(len(columnWidths))])

    # Construct table to be printed
    table = []
    for row in array:
        rowList = []
        for idx, col in enumerate(row):
            rowList.append(str(col).ljust(columnWidths[idx]))
        table.append(' '.join(rowList))

    # Return table
    return '\n'.join(table)

# Print the relevant output for given output_type
def handle_output(results, output_type):
    if not output_type == 'QUIET':
        if not output_type == 'JSON':
            table = generate_table_result(logger, results, output_type == 'TABLE-NO-HEADER')
            if table:
                print(table)
        else:
            try:
                _ = (entry for entry in results)
                json_str = json.dumps(results)
                if json_str:
                    print(json_str)
            except:
                logger.debug('Output is not a JSON, and then cannot be printed with --output=JSON parameter.')
                pass

# Main function of Lexicon.
def main():
    # Dynamically determine all the providers available and gather command line arguments.
    parsed_args = generate_cli_main_parser().parse_args()

    log_level = logging.getLevelName(parsed_args.log_level)
    logging.basicConfig(stream=sys.stdout, level=log_level, format='%(message)s')
    logger.debug('Arguments: %s', parsed_args)

    # In the CLI context, will get configuration interactively:
    #   * from the command line
    #   * from the environment variables
    #   * from lexicon configuration files in working directory
    config = ConfigResolver()
    config.with_args(parsed_args).with_env().with_config_dir(os.getcwd())

    client = Client(config)
    
    results = client.execute()

    handle_output(results, parsed_args.output)

if __name__ == '__main__':
    main()



from __future__ import absolute_import

import logging


from lexicon.providers.base import Provider as BaseProvider

try:
    # this module uses the optional `PyNamecheap` library from PyPi
    import namecheap # optional dep
except ImportError:
    pass

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['namecheap.com']

def ProviderParser(subparser):
    subparser.add_argument(
        '--auth-token',
        help='specify api token for authentication'
    )
    
    # earlier versions of the API expected the email address here
    # now they appear to want the username.
    subparser.add_argument(
        '--auth-username',
        help='specify username for authentication'
    )

    # FIXME What is the client IP used for?
    # this doesn't seem to be used for anything, but is required by their API
    # namecheap requires API requests to come from whitelisted domains, and this
    # is probably updated with the actual IP on their end.
    subparser.add_argument(
        '--auth-client-ip',
        help='Client IP address to send to Namecheap API calls',
        default='127.0.0.1'
    )
    subparser.add_argument(
        '--auth-sandbox',
        help='Whether to use the sandbox server',
        action='store_true'
    )


class Provider(BaseProvider):
    """
    ========================================================================
    WARNING
    ========================================================================
    The Namecheap API does not add/update/delete a host but "gets" and
    "sets" ALL hosts at once (a complete replacement). In their comments on
    the API docs (https://www.namecheap.com/support/api/methods/domains-dns/set-hosts.aspx)
    it appears that not all host types are handled by their system.

    Their API only handles `[A, AAAA, CNAME, MX, MXE, TXT, URL, URL301, FRAME]`

    If you have SRV record, it may get lost. Also records configured as
    `A + DDNS` on their control panel will be downgrated to `A` records.
    """

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.client = namecheap.Api(
            ApiUser=self._get_provider_option('auth_username') or '',
            ApiKey=self._get_provider_option('auth_token') or '',
            UserName=self._get_provider_option('auth_username') or '',
            ClientIP=self._get_provider_option('auth_client_ip') or '',
            sandbox=self._get_provider_option('auth_sandbox') or False,
            debug=False,
        )
        self.domain = self.domain
        self.domain_id = None

    def authenticate(self):
        """
        The Namecheap API is a little difficult to work with.
        Originally this method called PyNamecheap's `domains_getList`, which is
        connected to an API endpoint that only lists domains registered under
        the authenticating account. However, an authenticating Namecheap user
        may be permissioned to manage the DNS of additional domains. Namecheap's
        API does not offer a way to list these domains.

        This approach to detecting permissioned relies on some implementation
        details of the Namecheap API and the PyNamecheap module:

        * If the user does not own the domain, or is not permissioned to manage
          it in any way, Namecheap will return an error status, which
          PyNamecheap will instantly catch and raise.
        * If a non-error repsonse is returned, the XML payload is analyzed.
          If the user owns the domain it immediately returns valid. Otherwise
          we look for "All" Modification rights, or the hosts-edit permission.

        This is not feature complete and most-likely misses multiple scenarios
        where:
        * a user is privileged to manage the domain DNS, but via another "right"
        * a user is privileged to manage the domain, but DNS is not configured

        Important Note:
        * the Namecheap API has inconsistent use of capitalization with strings
          and a case-insensitive match should be made. e.g. the following appear
          in a payload: `False` and 'false', 'OK' and 'Ok'.

        TODO:
        * check payload for PremiumDNS
          <PremiumDnsSubscription>
            <IsActive>false</IsActive>
          </PremiumDnsSubscription>
        * check payload for other types of DNS
          <DnsDetails ProviderType="FREE" IsUsingOurDNS="true" HostCount="5" EmailType="No Email Service" DynamicDNSStatus="false" IsFailover="false">
            <Nameserver>dns1.registrar-servers.com</Nameserver>
            <Nameserver>dns2.registrar-servers.com</Nameserver>
          </DnsDetails>
        """
        extra_payload = {'DomainName': self.domain, }

        try:
            xml = self.client._call('namecheap.domains.getInfo', extra_payload)
        except namecheap.ApiError as err:
            # this will happen if there is an API connection error
            # OR if the user is not permissioned to manage this domain
            # OR the API request came from a not whitelisted IP
            # we should print the error, so people know how to correct it.
            raise Exception('Authentication failed: `%s`' % str(err))

        xpath = './/{%(ns)s}CommandResponse/{%(ns)s}DomainGetInfoResult' % {'ns': namecheap.NAMESPACE}
        domain_info = xml.find(xpath)

        def _check_hosts_permission():
            # this shouldn't happen
            if domain_info is None:
                return False

            # do they own the domain?
            if domain_info.attrib['IsOwner'].lower() == 'true':
                return True

            # look for rights
            xpath_alt = './/{%(ns)s}CommandResponse/{%(ns)s}DomainGetInfoResult/{%(ns)s}Modificationrights' % {'ns': namecheap.NAMESPACE}
            rights_info = xml.find(xpath_alt)
            if rights_info is None:
                return False

            # all rights
            if rights_info.attrib['All'].lower() == 'true':
                return True

            for right in rights_info.getchildren():
                if right.attrib['Type'].lower() == 'hosts':
                    if right.text.lower() == 'ok':
                        return True
                    else:
                        # we're only looking at hosts, so we can exit now
                        return False

            return None

        permissioned = _check_hosts_permission()
        if not permissioned:
            raise Exception('The domain {} is not controlled by this Namecheap '
                            'account'.format(self.domain))

        # FIXME What is this for?
        self.domain_id = self.domain

    def option_ttl(self):
        """
        Parse the `options` for a TTL.
        Used as a distinct function for documentation.
        via https://www.namecheap.com/support/api/methods/domains-dns/set-hosts.aspx
            Time to live for all record types.
            Possible values: any value between 60 to 60000
            Default Value: 1800
        if a TTL is submitted on the commandline:
            this will parse, but not validate it
        if no TTL is submitted:
            it will be ignored, allowing the server-side default
        Please note:
            Namecheap appears to have an internal cache on records; even with a
            short TTL of 60 seconds, it may take 120 seconds for their DNS to
            propagate.
        """
        return self._get_lexicon_option('ttl')
        
    # Create record. If record already exists with the same content, do nothing
    def create_record(self, type, name, content):
        record = {
            # required
            'Type': type,
            'Name': self._relative_name(name),
            'Address': content,
        }
        # inject the ttl if specified
        option_ttl = self.option_ttl()
        if option_ttl:
            record['TTL'] = option_ttl
        # logger.debug('create_record: %s', 'id' in payload)
        # return 'id' in payload
        self.client.domains_dns_addHost(self.domain, record)
        return True

    # List all records. Return an empty list if no records found.
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is
    # received.
    def list_records(self, type=None, name=None, content=None, id=None):
        records = []
        raw_records = self.client.domains_dns_getHosts(self.domain)
        for record in raw_records:
            records.append(self._convert_to_lexicon(record))

        if id:
            records = [record for record in records if record['id'] == id]
        if type:
            records = [record for record in records if record['type'] == type]
        if name:
            if name.endswith('.'):
                name = name[:-1]
            records = [record for record in records if name in record['name'] ]
        if content:
            records = [record for record in records if record['content'].lower() == content.lower()]

        logger.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):
        # Delete record if it exists
        self.delete_record(identifier, type, name, content)
        return self.create_record(type, name, content)

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        records = self.list_records(type=type, name=name, content=content, id=identifier)
        for record in records:
            self.client.domains_dns_delHost(self.domain, self._convert_to_namecheap(record))
        return True

    def _convert_to_namecheap(self, record):
        """ converts from lexicon format record to namecheap format record,
        suitable to sending through the api to namecheap"""

        name = record['name']
        if name.endswith('.'):
            name = name[:-1]

        short_name = name[:name.find(self.domain)-1]
        processed_record = {
            'Type': record['type'],
            'Name': short_name,
            'TTL': record['ttl'],
            'Address': record['content'],
            'HostId': record['id']
        }

        return processed_record

    def _convert_to_lexicon(self, record):
        """ converts from namecheap raw record format to lexicon format record
        """

        name = record['Name']
        if self.domain not in name:
            name = "{}.{}".format(name,self.domain)

        processed_record = {
            'type': record['Type'],
            'name': '{0}.{1}'.format(record['Name'], self.domain),
            'ttl': record['TTL'],
            'content': record['Address'],
            'id': record['HostId']
        }

        return processed_record



"""Rackspace provider implementation"""

from __future__ import absolute_import

import json
import logging
import time

import requests

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['rackspacecloud.com']

def _async_request_completed(payload):
    """Looks into an async response payload to see if the requested job has finished."""
    if payload['status'] == 'COMPLETED':
        return True
    if payload['status'] == 'ERROR':
        return True
    return False

def ProviderParser(subparser):
    subparser.add_argument("--auth-account", help="specify account number for authentication")
    subparser.add_argument("--auth-username", help="specify username for authentication. Only used if --auth-token is empty.")
    subparser.add_argument("--auth-api-key", help="specify api key for authentication. Only used if --auth-token is empty.")
    subparser.add_argument("--auth-token", help="specify token for authentication. If empty, the username and api key will be used to create a token.")
    subparser.add_argument("--sleep-time", type=float, default=1, help="number of seconds to wait between update requests.")

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://dns.api.rackspacecloud.com/v1.0'
        self.auth_api_endpoint = 'https://identity.api.rackspacecloud.com/v2.0'

    def authenticate(self):
        self._auth_token = self._get_provider_option('auth_token')
        if not self._auth_token:
            auth_response = self._auth_request('POST', '/tokens', {
                'auth': {
                    'RAX-KSKEY:apiKeyCredentials': {
                        'username': self._get_provider_option('auth_username'),
                        'apiKey': self._get_provider_option('auth_api_key')
                    }
                }
            })
            self._auth_token = auth_response['access']['token']['id']

        payload = self._get('/domains', {
            'name': self.domain
        })

        if not payload['domains']:
            raise Exception('No domain found')
        if len(payload['domains']) > 1:
            raise Exception('Too many domains found. This should not happen')

        self.domain_id = payload['domains'][0]['id']


    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, type, name, content):
        data = {'records': [{'type': type, 'name': self._full_name(name), 'data': content}]}
        if self._get_lexicon_option('ttl'):
            data['records'][0]['ttl'] = self._get_lexicon_option('ttl')

        try:
            payload = self._post_and_wait('/domains/{0}/records'.format(self.domain_id), data)
        except Exception as e:
            if str(e).startswith('Record is a duplicate of another record'):
                return self.update_record(None, type, name, content)
            raise e

        success = len(payload['records']) > 0
        logger.debug('create_record: %s', success)
        return success

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        params = {'per_page': 100}
        if type:
            params['type'] = type
        if name:
            params['name'] = self._full_name(name)
        # Sending the data filter to the Rackspace DNS API results in a 503 error
        # if content:
        #     params['data'] = content

        payload = self._get('/domains/{0}/records'.format(self.domain_id), params)

        records = list(payload['records'])
        if content:
            records = [record for record in records if record['data'] == content]
        records = [{
            'type': record['type'],
            'name': record['name'],
            'ttl': record['ttl'],
            'content': record['data'],
            'id': record['id']
        } for record in records]

        logger.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def update_record(self, identifier, type=None, name=None, content=None):
        data = {}
        if type:
            data['type'] = type
        if name:
            data['name'] = self._full_name(name)
        if content:
            data['data'] = content
        if self._get_lexicon_option('ttl'):
            data['ttl'] = self._get_lexicon_option('ttl')

        if identifier is None:
            records = self.list_records(type, name)
            if len(records) < 1:
                raise Exception('Unable to find record to modify: ' + name)
            identifier = records[0]['id']

        self._put_and_wait('/domains/{0}/records/{1}'.format(self.domain_id, identifier), data)

        # If it didn't raise from the http status code, then we're good
        logger.debug('update_record: %s', identifier)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_record_id = []
        if not identifier:
            records = self.list_records(type, name, content)
            delete_record_id = [record['id'] for record in records]
        else:
            delete_record_id.append(identifier)

        logger.debug('delete_records: %s', delete_record_id)

        for record_id in delete_record_id:
            payload = self._delete_and_wait(
                '/domains/{0}/records/{1}'.format(self.domain_id, record_id)
            )

        # If it didn't raise from the http status code, then we're good
        success = True
        logger.debug('delete_record: %s', success)
        return success


    # Helpers
    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        full_url = (self.api_endpoint + '/{0}' + url).format(self._get_provider_option('auth_account'))
        r = requests.request(action, full_url, params=query_params,
                             data=json.dumps(data),
                             headers={
                                 'X-Auth-Token': self._get_provider_option('auth_token'),
                                 'Content-Type': 'application/json'
                             })
        r.raise_for_status()  # if the request fails for any reason, throw an error.
        return r.json()

    # Non-GET requests to the Rackspace CloudDNS API are asynchronous
    def _request_and_wait(self, action='POST', url='/', data=None, query_params=None):
        result = self._request(action, url, data, query_params)

        sleep_time = self._get_provider_option('sleep_time') or '1'
        sleep_time = float(sleep_time)

        while not _async_request_completed(result):
            if sleep_time:
                time.sleep(sleep_time)
            result = self._update_response(result)

        if result['status'] == 'ERROR':
            raise Exception(result['error']['details'])

        if 'response' in result:
            return result['response']
        return None

    def _post_and_wait(self, url='/', data=None, query_params=None):
        return self._request_and_wait('POST', url, data, query_params)

    def _put_and_wait(self, url='/', data=None, query_params=None):
        return self._request_and_wait('PUT', url, data, query_params)

    def _delete_and_wait(self, url='/', data=None, query_params=None):
        return self._request_and_wait('DELETE', url, data, query_params)

    def _update_response(self, payload):
        r = requests.request('GET', payload['callbackUrl'], params={'showDetails': 'true'},
                             data={},
                             headers={
                                 'X-Auth-Token': self._get_provider_option('auth_token'),
                                 'Content-Type': 'application/json'
                             })

        r.raise_for_status()  # if the request fails for any reason, throw an error.
        return r.json()

    def _auth_request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}

        r = requests.request(action, self.auth_api_endpoint + url, params=query_params,
                             data=json.dumps(data),
                             headers={
                                 'Content-Type': 'application/json'
                             })
        r.raise_for_status()  # if the request fails for any reason, throw an error.
        return r.json()



from __future__ import absolute_import

import json
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['online.net']

def ProviderParser(subparser):
    subparser.add_argument("--auth-token", help="specify private api token")

def to_data(type, content):
    if type == "TXT":
        return '"{0}"'.format(content)
    else:
        return content

class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.zone_name = 'Zone Automatic Lexicon '
        self.passive_zone = None
        self.active_zone = None
        self.domain_id = self.domain
        self.api_endpoint = 'https://api.online.net/api/v1'

    def authenticate(self):
        self.init_zones()

    def list_zones(self):
        return self._get('/domain/{0}/version'.format(self.domain_id))

    def init_zones(self):
        # sets current zone version
        zone_name_a = self.zone_name + 'A'
        zone_name_b = self.zone_name + 'B'
        active_row = None
        passive_row = None
        for row in self.list_zones():
            if row['active'] == True:
                active_row = row
            elif row['name'] == zone_name_a or row['name'] == zone_name_b:
                passive_row = row

        if passive_row is None:
            passive_row = self._post('/domain/{0}/version'.format(self.domain_id), {
                'name': zone_name_b if active_row['name'] == zone_name_a else zone_name_a
            })

        self.active_zone = active_row['uuid_ref']
        self.passive_zone = passive_row['uuid_ref']
        self.update_passive_zone()


    def update_passive_zone(self):
        self._put(
            '/domain/{0}/version/{1}/zone_from_bind'.format(
                self.domain_id,
                self.passive_zone
            ),
            self.get_bind_zone()
        )

    def get_bind_zone(self):
        records = self.list_zone_records(self.active_zone)
         # then convert records to bind format
        bindStr = ''
        for record in records:
            bindStr = bindStr + '{0} {1} IN {2} {3}{4}\n'.format(
                record['name'] or '@',
                record['ttl'],
                record['type'],
                '{0} '.format(record['aux']) if 'aux' in record else '',
                record['data'] or ''
            )
        return bindStr

    def enable_zone(self):
        zone = self.passive_zone
        if zone is None:
            raise Exception("Could not enable uninitialized passive_zone")
        payload = self._patch('/domain/{0}/version/{1}/enable'.format(
            self.domain_id,
            zone
        ))
        self.passive_zone = self.active_zone
        self.active_zone = zone
        self.update_passive_zone()


    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, type, name, content):
        try:
            record = self.find_record(type, name, content)
            if record is not None:
                return True

            record = {
                'name': self._fqdn_name(name),
                'type': type,
                'data': to_data(type, content),
                'priority': self._get_lexicon_option('priority') or '',
                'ttl': self._get_lexicon_option('ttl') or ''
            }

            payload = self._post(
                '/domain/{0}/version/{1}/zone'.format(
                    self.domain_id,
                    self.passive_zone
                ),
                record
            )
        except Exception as e:
            logger.debug(e)
            return False

        self.enable_zone()
        logger.debug('create_record: %s', True)
        return True

    def find_zone_records(self, zone, type=None, name=None, content=None):
        records = []
        for record in self.list_zone_records(zone):
            processed_record = {
                'id': record['id'],
                'type': record['type'],
                'name': self._full_name(record['name']),
                'ttl': record['ttl'],
                'content': record['data'],
                'priority': record['aux'] if 'aux' in record else ''
            }
            records.append(self._clean_TXT_record(processed_record))

        if type:
            records = [record for record in records if record['type'] == type]
        if name:
            fullName = self._full_name(name)
            records = [record for record in records if record['name'] == fullName]
        if content:
            records = [record for record in records if record['content'] == content]

        logger.debug('list_records: %s', records)
        return records

    def list_zone_records(self, zone_id):
        return self._get('/domain/{0}/version/{1}/zone'.format(self.domain_id, zone_id))

    def list_records(self, type=None, name=None, content=None):
        return self.find_zone_records(self.passive_zone, type, name, content)

    def find_record(self, type=None, name=None, content=None):
        record = None
        records = self.list_records(type, name, content)
        if len(records) < 1:
            return None
        else:
            return records[0]


    # Create or update a record.
    def update_record(self, id, type=None, name=None, content=None):
        record = self.find_record(type, name)
        if record is None:
            logger.debug("cannot find record to update: %s %s %s", id, type, name)
            return True
        if type:
            record['type'] = type
        if name:
            record['name'] = self._fqdn_name(name)
        if content:
            record['data'] = to_data(type, content)
        if self._get_lexicon_option('ttl'):
            record['ttl'] = self._get_lexicon_option('ttl')
        # it is weird that 'aux' becomes 'priority' in online's api
        if self._get_lexicon_option('priority'):
            record['priority'] = self._get_lexicon_option('priority')

        if id is None:
            id = record['id']

        record.pop('id')

        try:
            payload = self._patch('/domain/{0}/version/{1}/zone/{2}'.format(
                self.domain_id,
                self.passive_zone,
                id
            ), record)

        except Exception as e:
            logger.debug(e)
            return False

        self.enable_zone()
        # If it didn't raise from the http status code, then we're good
        logger.debug('update_record: %s', id)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, id=None, type=None, name=None, content=None):
        records = self.list_records(type, name, content)
        if len(records) == 0:
            logger.debug("Cannot find records %s %s %s", type, name, content)
            return False
        logger.debug('delete_records: %s records found', len(records))
        try:
            for record in records:
                payload = self._delete('/domain/{0}/version/{1}/zone/{2}'.format(
                    self.domain_id,
                    self.passive_zone,
                    record['id']
                ))
        except Exception as e:
            logger.debug(e)
            return False

        self.enable_zone()
        # is always True at this point, if a non 200 response is returned an error is raised.
        logger.debug('delete_record: %s', True)
        return True

    def _patch(self, url='/', data=None, query_params=None):
        return self._request('PATCH', url, data=data, query_params=query_params)

    # Helpers
    def _request(self, action='GET',  url='/', data=None, query_params=None):
        if query_params is None:
            query_params = {}

        headers = {
            'Accept': 'application/json',
            'Authorization': 'Bearer {0}'.format(self._get_provider_option('auth_token'))
        }
        if data is not None:
            if type(data) is str:
                headers['Content-Type'] = 'text/plain';
            else:
                headers['Content-Type'] = 'application/json';
                data = json.dumps(data)

        r = requests.request(
            action,
            self.api_endpoint + url,
            params=query_params,
            data=data,
            headers=headers
        )
        r.raise_for_status()  # if the request fails for any reason, throw an error.

        return r.text and r.json() or ''



from __future__ import absolute_import

import json
import logging
import re
import base64
import copy

import requests
from requests.auth import HTTPBasicAuth

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['gehirn.jp']

def ProviderParser(subparser):
    subparser.add_argument(
        "--auth-token", help="specify access token for authentication")
    subparser.add_argument(
        "--auth-secret", help="specify access secret for authentication")


BUILD_FORMATS = {
    "A": "{address}",
    "AAAA": "{address}",
    "CNAME": "{cname}",
    "TXT": "{data}",
    "NS": "{nsdname}",
    "MX": "{prio} {exchange}",
    "SRV": "{prio} {weight} {port} {target}",
}

FORMAT_RE = {
    "A": re.compile("(?P<address>.+)"),
    "AAAA": re.compile("(?P<address>.+)"),
    "CNAME": re.compile("(?P<cname>.+)"),
    "TXT": re.compile("(?P<data>.+)"),
    "NS": re.compile("(?P<nsdname>.+)"),
    "MX": re.compile("(?P<prio>\d+)\s+(?P<exchange>.+)"),
    "SRV": re.compile("(?P<prio>\d+)\s+(?P<weight>\d+)\s+(?P<port>\d+)\s+(?P<target>.+)"),
}


class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.version_id = None
        self.api_endpoint = 'https://api.gis.gehirn.jp/dns/v1'

    def authenticate(self):
        payload = self._get('/zones')

        domains = [item for item in payload if item['name']
                   == self.domain]
        if not domains:
            raise Exception('No domain found')

        self.domain_id = domains[0]["id"]
        self.version_id = domains[0]["current_version_id"]

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, type, name, content):
        name = self._full_name(name)
        r = self._parse_content(type, content)

        record = None
        records = self._get_records(type=type, name=name)
        if len(records) == 0:
            record = {
                'type': type,
                'name': name,
                'enable_alias': False,
                'ttl': self._get_lexicon_option('ttl'),
                'records': [],
            }
        else:
            record = records[0]

        if r in record["records"]:
            logger.debug('create_record: %s', True)
            return True

        record["records"].append(r)
        self._update_record(record)
        logger.debug('create_record: %s', True)
        return True

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        records = []
        if name:
            name = self._full_name(name)
        for record in self._get_records(type=type, name=name):
            for i, r in enumerate(record["records"]):
                c = self._build_content(record['type'], r)
                processed_record = {
                    'type': record['type'],
                    'name': record['name'].rstrip("."),
                    'ttl': record['ttl'],
                    'content': c,
                    'id': "{}.{}".format(record["id"], base64.b64encode(c.encode("utf-8")).decode("ascii")),
                }
                self._parse_content(
                    record['type'], processed_record["content"])
                records.append(processed_record)

        if content:
            records = [
                record for record in records if record['content'] == content]

        logger.debug('list_records: %s', records)
        return records

    # Create or update a record.
    def update_record(self, identifier=None, type=None, name=None, content=None):

        if name:
            name = self._full_name(name)

        record = None
        if not identifier:
            if not (type and name and content):
                raise Exception("type, name and content must be specified.")
                r = self._parse_content(type, content)

            records = self._get_records(type=type, name=name)

            if not records:
                self.create_record(type=type, name=name, content=content)
                logger.debug('update_record: %s', True)
                return True

            record = {
                'id': records[0]["id"],
                'type': type,
                'name': name,
                'enable_alias': False,
                'ttl': self._get_lexicon_option('ttl'),
                'records': [self._parse_content(type, content)],
            }

        else:
            # with identifier
            records = self._get_records(identifier=identifier)
            if not records:
                raise Exception('Record identifier could not be found.')

            record = records[0]

            if "." in identifier:
                # modify single record
                self.delete_record(identifier=identifier)
                self.create_record(
                    type=type or record["type"],
                    name=name or record["name"],
                    content=content
                )
                logger.debug('update_record: %s', True)
                return True
            else:
                # update entire record
                if type:
                    record["type"] = type
                if name:
                    record["name"] = name
                record["ttl"] = self._get_lexicon_option('ttl')
                if content:
                    record["records"] = [
                        self._parse_content(record["type"], content)]

        self._update_record(record)
        logger.debug('update_record: %s', True)
        return True

    # Delete an existing record.
    # If record does not exist, do nothing.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        if identifier:
            if "." not in identifier:
                # delete entire record
                path = '/zones/{}/versions/{}/records/{}'.format(
                    self.domain_id, self.version_id, identifier,
                )
                self._delete(path)
                logger.debug('delete_record: %s', True)
                return True

            record_identifier = identifier.split(".")[1]

            records = self._get_records(identifier=identifier)
            if not records:
                raise Exception('Record identifier could not be found.')

            record = records[0]
            for index, r in enumerate(record["records"]):
                target_content = self._build_content(record['type'], r)
                target_identifier = base64.b64encode(
                    target_content.encode("utf-8")).decode("ascii")

                if target_identifier == record_identifier:
                    del record["records"][index]
                if len(record["records"]) == 0:
                    # delete entire record
                    path = '/zones/{}/versions/{}/records/{}'.format(
                        self.domain_id, self.version_id, record['id'],
                    )
                    self._delete(path)
                else:
                    self._update_record(record)

                logger.debug('delete_record: %s', True)
                return True
            else:
                raise Exception('Record identifier could not be found.')

        r = None
        if name is not None:
            name = self._full_name(name)
        if content is not None:
            content = self._bind_format_target(type, content)
            r = self._parse_content(type, content)

        records = self._get_records(type=type, name=name)

        for record in records:
            if r and r in record["records"]:
                record["records"].remove(r)
                if len(record["records"]):
                    self._update_record(record)
                    continue

            path = '/zones/{}/versions/{}/records/{}'.format(
                self.domain_id, self.version_id, record["id"],
            )
            self._delete(path)

        logger.debug('delete_record: %s', True)
        return True

    # Helpers
    def _full_name(self, name):
        name = super(Provider, self)._full_name(name)
        if not name.endswith("."):
            name += "."
        return name

    def _bind_format_target(self, type, target):
        if type == "CNAME" and not target.endswith("."):
            target += "."
        return target

    def _filter_records(self, records, identifier=None, type=None, name=None):
        filtered_records = []

        if identifier:
            identifier = identifier.split(".")[0]

        for record in records:
            if type and record['type'] != type:
                continue
            if name and record['name'] != name:
                continue
            if identifier and record['id'] != identifier:
                continue
            filtered_records.append(record)
        return filtered_records

    def _get_records(self, identifier=None, type=None, name=None):
        path = '/zones/{}/versions/{}/records'.format(
            self.domain_id, self.version_id)
        return self._filter_records(self._get(path), identifier=identifier, type=type, name=name)

    def _update_record(self, record):
        if record.get("id"):
            # PUT
            path = '/zones/{}/versions/{}/records/{}'.format(
                self.domain_id, self.version_id, record["id"],
            )
            return self._put(path, record)

        # POST
        path = '/zones/{}/versions/{}/records'.format(
            self.domain_id, self.version_id,
        )
        return self._post(path, record)

    def _build_content(self, type, record):
        return BUILD_FORMATS[type].format(**record)

    def _parse_content(self, type, content):
        return FORMAT_RE[type].match(content).groupdict()

    def _request(self, action='GET',  url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        default_headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json',
        }
        default_auth = HTTPBasicAuth(
            self._get_provider_option('auth_token'), self._get_provider_option('auth_secret'))

        query_string = ""
        if query_params:
            query_string = json.dumps(query_params)

        r = requests.request(action, self.api_endpoint + url, params=query_string,
                             data=json.dumps(data) if data else None,
                             headers=default_headers,
                             auth=default_auth)
        try:
            # if the request fails for any reason, throw an error.
            r.raise_for_status()
        except:
            logger.error(r.text)
            raise
        return r.json()



from __future__ import absolute_import

from builtins import object

from lexicon.config import ConfigResolver
from lexicon.config import legacy_config_resolver


class Provider(object):

    """
    This is the base class for all lexicon Providers. It provides common functionality and ensures that all implemented
    Providers follow a standard ducktype. All standardized options will be provided here as defaults, but can be overwritten
    by environmental variables and cli arguments.

    Common options are:

    action
    domain
    type
    name
    content
    ttl
    priority
    identifier

    The provider_env_cli_options will also contain any Provider specific options:

    auth_username
    auth_token
    auth_password
    ...

    :param provider_env_cli_options: is a ConfigResolver object that contains all the options for this provider, merged from CLI and Env variables.
    """
    def __init__(self, config):
        if not isinstance(config, ConfigResolver):
            # If config is a plain dict, we are in a legacy situation.
            # To protect the Provider API, the legacy dict is handled in a
            # correctly defined ConfigResolver.
            self.config = legacy_config_resolver(config)
        else:
            self.config = config

        # Default ttl
        self.config.with_dict({'ttl': 3600})

        self.provider_name = self.config.resolve('lexicon:provider_name') or self.config.resolve('lexicon:provider')
        self.domain = self.config.resolve('lexicon:domain')
        self.domain_id = None

    # Authenticate against provider,
    # Make any requests required to get the domain's id for this provider, so it can be used in subsequent calls.
    # Should throw an error if authentication fails for any reason, of if the domain does not exist.
    def authenticate(self):
        raise NotImplementedError("Providers should implement this!")

    # Create record. If record already exists with the same content, do nothing'
    def create_record(self, type, name, content):
        raise NotImplementedError("Providers should implement this!")

    # List all records. Return an empty list if no records found
    # type, name and content are used to filter records.
    # If possible filter during the query, otherwise filter after response is received.
    def list_records(self, type=None, name=None, content=None):
        raise NotImplementedError("Providers should implement this!")

    # Update a record. Identifier must be specified.
    def update_record(self, identifier, type=None, name=None, content=None):
        raise NotImplementedError("Providers should implement this!")

    # Delete an existing record.
    # If record does not exist, do nothing.
    # If an identifier is specified, use it, otherwise do a lookup using type, name and content.
    def delete_record(self, identifier=None, type=None, name=None, content=None):
        raise NotImplementedError("Providers should implement this!")

    #Helpers
    def _request(self, action='GET',  url='/', data=None, query_params=None):
        raise NotImplementedError("Providers should implement this!")

    def _get(self, url='/', query_params=None):
        return self._request('GET', url, query_params=query_params)

    def _post(self, url='/', data=None, query_params=None):
        return self._request('POST', url, data=data, query_params=query_params)

    def _put(self, url='/', data=None, query_params=None):
        return self._request('PUT', url, data=data, query_params=query_params)

    def _delete(self, url='/', query_params=None):
        return self._request('DELETE', url, query_params=query_params)

    def _fqdn_name(self, record_name):
        record_name = record_name.rstrip('.') # strip trailing period from fqdn if present
        #check if the record_name is fully specified
        if not record_name.endswith(self.domain):
            record_name = "{0}.{1}".format(record_name, self.domain)
        return "{0}.".format(record_name) #return the fqdn name

    def _full_name(self, record_name):
        record_name = record_name.rstrip('.') # strip trailing period from fqdn if present
        #check if the record_name is fully specified
        if not record_name.endswith(self.domain):
            record_name = "{0}.{1}".format(record_name, self.domain)
        return record_name

    def _relative_name(self, record_name):
        record_name = record_name.rstrip('.') # strip trailing period from fqdn if present
        #check if the record_name is fully specified
        if record_name.endswith(self.domain):
            record_name = record_name[:-len(self.domain)]
            record_name = record_name.rstrip('.')
        return record_name

    def _clean_TXT_record(self, record):
        if record['type'] == 'TXT':
            # some providers have quotes around the TXT records, so we're going to remove those extra quotes
            record['content'] = record['content'][1:-1]
        return record

    def _get_lexicon_option(self, option):
        return self.config.resolve('lexicon:{0}'.format(option))

    def _get_provider_option(self, option):
        return self.config.resolve('lexicon:{0}:{1}'.format(self.provider_name, option))



from __future__ import absolute_import

import requests
import json
import logging

from lexicon.providers.base import Provider as BaseProvider

LOGGER = logging.getLogger(__name__)

NAMESERVER_DOMAINS = ['zeit.world']

def ProviderParser(subparser):
    subparser.description = '''
        Zeit Provider requires a token to access its API.
        You can generate one for your account on the following URL:
        https://zeit.co/account/tokens'''
    subparser.add_argument('--auth-token', help='specify your API token')

# Implements the DNS Zeit provider.
# The API is quite simple: you can list all records, add one record or delete one record.
#   - list is pretty straightforward: we get all records then filter for given parameters,
#   - add uses directly the API to add a new record without any added complexity,
#   - delete uses list + delete: we get the list of all records, filter on the given parameters and delete record by id,
#   - update uses list + delete + add: we get the list of all records, find record for given identifier, then insert a new record and delete the old record.
class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        self.api_endpoint = 'https://api.zeit.co/v2/domains'

    def authenticate(self):
        result = self._get('/{0}'.format(self.domain))

        if not result['uid']:
            raise Exception('Error, domain {0} not found'.format(self.domain))

        self.domain_id = result['uid']

    def list_records(self, type=None, name=None, content=None):
        result = self._get('/{0}/records'.format(self.domain))

        raw_records = result['records']
        if type:
            raw_records = [raw_record for raw_record in raw_records if raw_record['type'] == type]
        if name:
            raw_records = [raw_record for raw_record in raw_records if raw_record['name'] == self._relative_name(name)]
        if content:
            raw_records = [raw_record for raw_record in raw_records if raw_record['value'] == content]
        
        records = []
        for raw_record in raw_records:
            records.append({
                'id': raw_record['id'],
                'type': raw_record['type'],
                'name': self._full_name(raw_record['name']),
                'content': raw_record['value']
            })

        LOGGER.debug('list_records: %s', records)

        return records

    def create_record(self, type, name, content):
        # We ignore creation if a record already exists for given type/name/content
        records = self.list_records(type, name, content)
        if records:
            LOGGER.debug('create_record (ignored, duplicate): %s', records[0]['id'])
            return True
        
        data = {
            'type': type,
            'name': self._relative_name(name),
            'value': content
        }

        result = self._post('/{0}/records'.format(self.domain), data)

        if not result['uid']:
            raise Exception('Error occured when inserting the new record.')

        LOGGER.debug('create_record: %s', result['uid'])

        return True

    def update_record(self, identifier, type=None, name=None, content=None):
        # Zeit do not allow to update a record, only add or remove.
        # So we get the corresponding record, dump or update its content and insert it as a new record.
        # Then we remove the old record.
        records = []
        if identifier:
            records = self.list_records()
            records = [record for record in records if record['id'] == identifier]
        else:
            records = self.list_records(type, name)
        
        if not records:
            raise Exception('No record found for identifer: {0}'.format(identifier))

        if len(records) > 1:
            LOGGER.warn('Multiple records have been found for given parameters. Only first one will be updated (id: {0})'.format(records[0]['id']))

        data = {
            'type': type,
            'name': self._relative_name(name),
            'value': content
        }

        if not type:
            data['type'] = records[0]['type']
        if not name:
            data['name'] = self._relative_name(records[0]['name'])
        if not content:
            data['value'] = records[0]['content']

        result = self._post('/{0}/records'.format(self.domain), data)
        self._delete('/{0}/records/{1}'.format(self.domain, records[0]['id']))

        LOGGER.debug('update_record: %s => %s', records[0]['id'], result['uid'])

        return True

    def delete_record(self, identifier=None, type=None, name=None, content=None):
        delete_record_ids = []
        if not identifier:
            records = self.list_records(type, name, content)
            delete_record_ids = [record['id'] for record in records]
        else:
            delete_record_ids.append(identifier)

        LOGGER.debug('delete_records: %s', delete_record_ids)

        for delete_record_id in delete_record_ids:
            self._delete('/{0}/records/{1}'.format(self.domain, delete_record_id))

        LOGGER.debug('delete_record: %s', True)

        return True

    def _request(self, action='GET',  url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        
        request = requests.request(action, self.api_endpoint + url, 
                                   params=query_params,
                                   data=json.dumps(data),
                                   headers={'Authorization': 'Bearer {0}'.format(self._get_provider_option('auth_token'))})

        request.raise_for_status()
        return request.json()



from __future__ import absolute_import

import logging
from xml.etree import ElementTree
import requests

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)

APIENTRYPOINT = {
        'zonomi': 'https://zonomi.com/app',
        'rimuhosting' : 'https://rimuhosting.com'
}

NAMESERVER_DOMAINS = ['zonomi.com']

# Lexicon Zonomi and Rimuhosting Provider
#
# Author: Juan Rossi, 2017
#
# Zonomi API Docs: https://zonomi.com/app/dns/dyndns.jsp
# Rimuhosting API Docs: https://rimuhosting.com/dns/dyndns.jsp
#
# Implementation notes:
# * Lots of tricks taken from the PowerDNS API
# * The Zonomi API does not assign a unique identifier to each record in the way
#   that Lexicon expects. We work around this by creating an ID based on the record
#   name, type and content, which when taken together are always unique
# * The  API has no notion of 'create a single record' or 'delete a single
#   record'. All operations are either 'replace the RRSet with this new set of records'
#   or 'delete all records for this name and type. Similarly, there is no notion of
#   'change the content of this record', because records are identified by their name,
#   type and content.


def ProviderParser(subparser):
    subparser.add_argument("--auth-token", help="specify token for authentication")
    subparser.add_argument("--auth-entrypoint", help="use Zonomi or Rimuhosting API", choices=[
        'zonomi', 'rimuhosting' ])


class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)
        self.domain_id = None
        if not self._get_provider_option('auth_token'):
            raise Exception('Error, application key is not defined')

        self.api_endpoint = APIENTRYPOINT.get('zonomi')

        if self._get_provider_option('auth_entrypoint'):
            self.api_endpoint = APIENTRYPOINT.get(self._get_provider_option('auth_entrypoint'))


    def authenticate(self):
    
        payload = self._get('/dns/dyndns.jsp', {
            'action' : 'QUERY',
            'name' : "**." + self.domain,
            'type' : 'SOA'
        })

        if payload.find('is_ok').text != 'OK:':
            raise Exception('Error with api {0}'.format(payload.find('is_ok').text))

        self.domain_id = self.domain


    def _make_identifier(self, type, name, content):
        return "{}/{}={}".format(type, self._full_name(name), content)


    def _parse_identifier(self, identifier):
        parts = identifier.split('/')
        type = parts[0]
        parts = parts[1].split('=')
        name = parts[0]
        content = "=".join(parts[1:])
        return type, name, content


    def create_record(self, type, name, content):
        request = {
                'action': 'SET', 
                'type': type, 
                'name': self.domain, 
                'value': content 
        }

        if name is not None:
            request['name'] = self._full_name(name)

        if self._get_lexicon_option('ttl'):
            request['ttl'] = self._get_lexicon_option('ttl')

        if self._get_lexicon_option('priority'):
            request['prio'] = self._get_lexicon_option('priority')

        payload = self._get('/dns/dyndns.jsp', request)

        if payload.find('is_ok').text != 'OK:':
            raise Exception('An error occurred: {0}'.format(payload.find('is_ok').text))

        logger.debug('create_record: %s', True)
        return True


    def list_records(self, type=None, name=None, content=None):
        records = []

        request = {
            'action' : 'QUERY',
            'name': "**." + self.domain
        }

        if type is not None:
            request['type'] = type
        if name is not None:
            request['name'] = self._full_name(name)
        if content is not None:
            request['value'] = content

        payload = self._get('/dns/dyndns.jsp', request)
        for rxml in payload.iter('record'):
            processed_record = {
                'type': rxml.attrib['type'],
                'name': rxml.attrib['name'],
                'content': rxml.attrib['content'],
                'id': self._make_identifier(rxml.attrib['type'],rxml.attrib['name'],rxml.attrib['content']),
                'ttl': rxml.attrib['ttl'].split()[0]
            }
            records.append(processed_record)
        logger.debug('list_records: %s', records)
        return records


    def delete_record(self, identifier=None, type=None, name=None, content=None):
        if identifier is not None:
            type, name, content = self._parse_identifier(identifier)

        request = {
            'action' : 'DELETE',
            'name': self.domain 
        }
        
        if type is not None:
            request['type'] = type
        if name is not None:
            request['name'] = self._full_name(name)
        if content is not None:
            request['value'] = content

        payload = self._get('/dns/dyndns.jsp', request)

        if payload.find('is_ok').text != 'OK:':
            raise Exception('An error occurred: {0}'.format(payload.find('is_ok').text))

        logger.debug('delete_record: %s', True)
        return True


    def update_record(self, identifier, type=None, name=None, content=None):
        self.delete_record(identifier)
        ttype, tname, tcontent = self._parse_identifier(identifier)
        request = {
                'action': 'SET',
                'type': ttype,
                'name': self._full_name(tname),
                'value': tcontent
        }

        if type is not None:
            request['type'] = type
        if name is not None:
            request['name'] = self._full_name(name)
        if content is not None:
            request['value'] = content
        if self._get_lexicon_option('ttl'):
            request['ttl'] = self._get_lexicon_option('ttl')
        if self._get_lexicon_option('priority'):
            request['prio'] = self._get_lexicon_option('priority')

        payload = self._get('/dns/dyndns.jsp', request)

        if payload.find('is_ok').text != 'OK:':
            raise Exception('An error occurred: {0}'.format(payload.find('is_ok').text))

        logger.debug('update_record: %s', True)
        return True

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        else:
            query_params['api_key'] = self._get_provider_option('auth_token')

        r = requests.request(action, self.api_endpoint + url, params=query_params)
        tree = ElementTree.ElementTree(ElementTree.fromstring(r.content))
        root = tree.getroot()
        if root.tag == 'error':
            raise Exception('An error occurred: {0}'.format(root.text))
        else:
            r.raise_for_status()
        return root



from __future__ import absolute_import

import json
import logging

import requests

from lexicon.providers.base import Provider as BaseProvider

logger = logging.getLogger(__name__)


# Lexicon PowerDNS Provider
#
# Author: Will Hughes, 2017
#
# API Docs: https://doc.powerdns.com/md/httpapi/api_spec/
#
# Implementation notes:
# * The PowerDNS API does not assign a unique identifier to each record in the way
#   that Lexicon expects. We work around this by creating an ID based on the record
#   name, type and content, which when taken together are always unique
# * The PowerDNS API has no notion of 'create a single record' or 'delete a single
#   record'. All operations are either 'replace the RRSet with this new set of records'
#   or 'delete all records for this name and type. Similarly, there is no notion of
#   'change the content of this record', because records are identified by their name,
#   type and content.
# * The API is very picky about the format of values used when creating records:
# ** CNAMEs must be fully qualified
# ** TXT, LOC records must be quoted
#   This is why the _clean_content and _unclean_content methods exist, to convert
#   back and forth between the format PowerDNS expects, and the format Lexicon uses

NAMESERVER_DOMAINS = []

def ProviderParser(subparser):
    subparser.add_argument("--auth-token", help="specify token for authentication")
    subparser.add_argument("--pdns-server", help="URI for PowerDNS server")
    subparser.add_argument("--pdns-server-id", help="Server ID to interact with")


class Provider(BaseProvider):

    def __init__(self, config):
        super(Provider, self).__init__(config)

        self.api_endpoint = self._get_provider_option('pdns_server')

        if self.api_endpoint.endswith('/'):
            self.api_endpoint = self.api_endpoint[:-1]

        if not self.api_endpoint.endswith("/api/v1"):
            self.api_endpoint += "/api/v1"

        self.server_id = self._get_provider_option('pdns_server_id')
        if self.server_id is None:
            self.server_id = 'localhost'

        self.api_endpoint += "/servers/" + self.server_id

        self.api_key = self._get_provider_option('auth_token')
        assert self.api_key is not None

        self._zone_data = None

    def zone_data(self):
        if self._zone_data is None:
            self._zone_data = self._get('/zones/' + self.domain).json()
        return self._zone_data

    def authenticate(self):
        self.zone_data()
        self.domain_id = self.domain

    def _make_identifier(self, type, name, content):
        return "{}/{}={}".format(type, name, content)

    def _parse_identifier(self, identifier):
        parts = identifier.split('/')
        type = parts[0]
        parts = parts[1].split('=')
        name = parts[0]
        content = "=".join(parts[1:])
        return type, name, content

    def list_records(self, type=None, name=None, content=None):
        records = []
        for rrset in self.zone_data()['rrsets']:
            if (name is None or self._fqdn_name(rrset['name']) == self._fqdn_name(name)) and (type is None or rrset['type'] == type):
                for record in rrset['records']:
                    if content is None or record['content'] == self._clean_content(type, content):
                        records.append({
                            'type': rrset['type'],
                            'name': self._full_name(rrset['name']),
                            'ttl': rrset['ttl'],
                            'content': self._unclean_content(rrset['type'], record['content']),
                            'id': self._make_identifier(rrset['type'], rrset['name'], record['content'])
                        })
        logger.debug('list_records: %s', records)
        return records

    def _clean_content(self, type, content):
        if type in ("TXT", "LOC"):
            if content[0] != '"':
                content = '"' + content
            if content[-1] != '"':
                content += '"'
        elif type == "CNAME":
            content = self._fqdn_name(content)
        return content

    def _unclean_content(self, type, content):
        if type in ("TXT", "LOC"):
            content = content.strip('"')
        elif type == "CNAME":
            content = self._full_name(content)
        return content

    def create_record(self, type, name, content):
        content = self._clean_content(type, content)
        for rrset in self.zone_data()['rrsets']:
            if rrset['name'] == name and rrset['type'] == type:
                update_data = rrset
                if 'comments' in update_data:
                    del update_data['comments']
                update_data['changetype'] = 'REPLACE'
                break
        else:
            update_data = {
                'name': name,
                'type': type,
                'records': [],
                'ttl': self._get_lexicon_option('ttl') or 600,
                'changetype': 'REPLACE'
            }

        for record in update_data['records']:
            if record['content'] == content:
                return True

        update_data['records'].append({
            'content': content,
            'disabled': False
        })

        update_data['name'] = self._fqdn_name(update_data['name'])

        request = {'rrsets': [update_data]}
        logger.debug('request: %s', request)

        self._patch('/zones/' + self.domain, data=request)
        self._zone_data = None
        return True

    def delete_record(self, identifier=None, type=None, name=None, content=None):
        if identifier is not None:
            type, name, content = self._parse_identifier(identifier)

        logger.debug("delete %s %s %s", type, name, content)
        if type is None or name is None:
            raise Exception("Must specify at least both type and name")

        for rrset in self.zone_data()['rrsets']:
            if rrset['type'] == type and self._fqdn_name(rrset['name']) == self._fqdn_name(name):
                update_data = rrset
                if 'comments' in update_data:
                    del update_data['comments']
                update_data['changetype'] = 'REPLACE'
                break
        else:
            return True

        new_records = []
        for record in update_data['records']:
            if content is None or self._unclean_content(type, record['content']) != self._unclean_content(type, content):
                new_records.append(record)

        update_data['name'] = self._fqdn_name(update_data['name'])
        update_data['records'] = new_records

        request = {'rrsets': [update_data]}
        logger.debug('request: %s', request)

        self._patch('/zones/' + self.domain, data=request)
        self._zone_data = None
        return True

    def update_record(self, identifier, type=None, name=None, content=None):
        self.delete_record(identifier)
        return self.create_record(type, name, content)

    def _patch(self, url='/', data=None, query_params=None):
        return self._request('PATCH', url, data=data, query_params=query_params)

    def _request(self, action='GET', url='/', data=None, query_params=None):
        if data is None:
            data = {}
        if query_params is None:
            query_params = {}
        r = requests.request(action, self.api_endpoint + url, params=query_params,
                             data=json.dumps(data),
                             headers={
                                 'X-API-Key': self.api_key,
                                 'Content-Type': 'application/json'
                             })
        logger.debug('response: %s', r.text)
        r.raise_for_status()
        return r
